#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections import OrderedDict
import json

import dateutil
import pytest
import responses

from datarobot.models import Deployment, ServiceStats, ServiceStatsOverTime
from datarobot.utils import from_api


class TestServiceStats(object):
    @pytest.fixture
    def minimal_response_data(self, deployment_data):
        return {
            "model_id": deployment_data["model"]["id"],
            "period": {
                "start": "2019-08-01T00:00:00.000000Z",
                "end": "2019-08-10T00:00:00.000000Z",
            },
            "metrics": {"totalPredictions": 1000, "executionTime": 50.27, "responseTime": None},
        }

    @pytest.fixture
    def no_stats_response_data(self, deployment_data):
        return {
            "model_id": deployment_data["model"]["id"],
            "period": {
                "start": "2019-08-01T00:00:00.000000Z",
                "end": "2019-08-10T00:00:00.000000Z",
            },
            "metrics": {
                "totalPredictions": 0,
                "totalRequests": 0,
                "slowRequests": 0,
                "executionTime": None,
                "responseTime": None,
                "userErrorRate": 0,
                "serverErrorRate": 0,
                "numConsumers": 0,
                "cacheHitRatio": 0,
                "medianLoad": 0,
                "peakLoad": 0,
            },
        }

    @pytest.fixture
    def extra_metrics_response_data(self, deployment_data):
        return {
            "model_id": deployment_data["model"]["id"],
            "period": {
                "start": "2019-08-01T00:00:00.000000Z",
                "end": "2019-08-10T00:00:00.000000Z",
            },
            "metrics": {
                "totalPredictions": 65122,
                "totalRequests": 32,
                "slowRequests": 8,
                "executionTime": 712,
                "responseTime": 828,
                "userErrorRate": 0.03,
                "serverErrorRate": 0.00002,
                "numConsumers": 1,
                "cacheHitRatio": 0.96875,
                "medianLoad": 2.456,
                "peakLoad": 15,
                "newIntMetric": 1,
                "newFloatMetric": 0.4675,
                "newNullableMetric": None,
            },
        }

    @pytest.fixture(
        params=["minimal_response_data", "no_stats_response_data", "extra_metrics_response_data"]
    )
    def response_data(self, request):
        return request.getfixturevalue(request.param)

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        url = "{}/deployments/{}/serviceStats/".format(unittest_endpoint, deployment_data["id"])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(response_data),
        )

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_service_stats(self, deployment_data, response_data):
        service_stats_objs = [
            ServiceStats.get(deployment_data["id"]),
            Deployment.get(deployment_data["id"]).get_service_stats(),
        ]
        for service_stats in service_stats_objs:
            assert service_stats.model_id == response_data["model_id"]

            expected_start_date = dateutil.parser.parse(response_data["period"]["start"])
            expected_end_date = dateutil.parser.parse(response_data["period"]["end"])
            assert service_stats.period["start"] == expected_start_date
            assert service_stats.period["end"] == expected_end_date

            assert sorted(service_stats.metrics.keys()) == sorted(response_data["metrics"].keys())
            for metric_name, metric in service_stats.metrics.items():
                expected_value = response_data["metrics"][metric_name]
                assert metric == expected_value
                assert service_stats[metric_name] == expected_value

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    @pytest.mark.parametrize(
        "kwarg, expected_query_string",
        [
            ({"model_id": "abc"}, "?modelId=abc"),
            ({"execution_time_quantile": 0.74}, "?executionTimeQuantile=0.74"),
            ({"response_time_quantile": 0.43}, "?responseTimeQuantile=0.43"),
            ({"slow_requests_threshold": 40}, "?slowRequestsThreshold=40"),
        ],
    )
    def test_params(self, deployment_data, kwarg, expected_query_string):
        """Test optional params are correctly forwarded to the API"""

        ServiceStats.get(deployment_data["id"], **kwarg)
        assert expected_query_string in responses.calls[0].request.url

        Deployment.get(deployment_data["id"]).get_service_stats(**kwarg)
        assert expected_query_string in responses.calls[0].request.url


class TestServiceStatsOverTime(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            "model_id": deployment_data["model"]["id"],
            "metric": "totalPredictions",
            "summary": {
                "period": {
                    "start": "2019-08-01T00:00:00.000000Z",
                    "end": "2019-08-04T00:00:00.000000Z",
                },
                "value": 2250,
            },
            "buckets": [
                {
                    "period": {
                        "start": "2019-08-01T00:00:00.000000Z",
                        "end": "2019-08-02T00:00:00.000000Z",
                    },
                    "value": 1250,
                },
                {
                    "period": {
                        "start": "2019-08-02T00:00:00.000000Z",
                        "end": "2019-08-03T00:00:00.000000Z",
                    },
                    "value": 1000,
                },
                {
                    "period": {
                        "start": "2019-08-03T00:00:00.000000Z",
                        "end": "2019-08-04T00:00:00.000000Z",
                    },
                    "value": 0,
                },
            ],
        }

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        deployment_id = deployment_data["id"]
        url = "{}/deployments/{}/serviceStatsOverTime/".format(unittest_endpoint, deployment_id)
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(response_data),
        )

    def assert_bucket(self, actual, expected):
        assert actual.keys() == expected.keys()

        actual_period = actual["period"]
        if actual_period:
            assert actual["period"].keys() == expected["period"].keys()
            assert actual_period["start"] == dateutil.parser.parse(expected["period"]["start"])
            assert actual_period["end"] == dateutil.parser.parse(expected["period"]["end"])
        else:
            assert actual_period is None

        assert actual["value"] == expected["value"]

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_service_stats_over_time(self, deployment_data, response_data):
        service_stats_over_time_objs = [
            ServiceStatsOverTime.get(deployment_data["id"]),
            Deployment.get(deployment_data["id"]).get_service_stats_over_time(),
        ]
        for service_stats_over_time in service_stats_over_time_objs:
            expected = from_api(response_data, keep_null_keys=True)
            assert service_stats_over_time.metric == expected["metric"]
            assert service_stats_over_time.model_id == expected["model_id"]

            self.assert_bucket(service_stats_over_time.summary, expected["summary"])
            for index, bucket in enumerate(service_stats_over_time.buckets):
                self.assert_bucket(bucket, expected["buckets"][index])

            expected_bucket_values = [
                (dateutil.parser.parse(bucket["period"]["start"]), bucket["value"])
                for bucket in expected["buckets"]
            ]
            assert service_stats_over_time.bucket_values == OrderedDict(expected_bucket_values)

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    @pytest.mark.parametrize(
        "kwarg, expected_query_string",
        [
            ({"metric": "LogLoss"}, "?metric=LogLoss"),
            ({"model_id": "abc"}, "?modelId=abc"),
            ({"bucket_size": "P1D"}, "?bucketSize=P1D"),
            ({"quantile": 0.74}, "?quantile=0.74"),
            ({"threshold": 0.43}, "?threshold=0.43"),
        ],
    )
    def test_params(self, deployment_data, kwarg, expected_query_string):
        """Test optional params are correctly forwarded to the API"""

        ServiceStatsOverTime.get(deployment_data["id"], **kwarg)
        assert expected_query_string in responses.calls[0].request.url

        Deployment.get(deployment_data["id"]).get_service_stats_over_time(**kwarg)
        assert expected_query_string in responses.calls[0].request.url



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json
from uuid import uuid4

import mock
import pandas as pd
import pytest
import responses
from six.moves.urllib_parse import parse_qs, urlparse
from trafaret import DataError

from datarobot import Feature, FeatureHistogram, ModelingFeature, Project
from datarobot.errors import InvalidUsageError
from datarobot.models.feature import (
    DatasetFeature,
    DatasetFeatureHistogram,
    FeatureLineage,
    InteractionFeature,
    MulticategoricalHistogram,
)
from datarobot.utils import camelize


@pytest.fixture
def list_features_json():
    return """
    [{
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "556cdfbb100d2b0e88585195",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }]
    """


@pytest.fixture
def feature_json():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "556cdfbb100d2b0e88585195",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "feature_lineage_id": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def feature_json_no_leakage():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "FALSE"
    }
    """


@pytest.fixture
def feature_json_leakage():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": true,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "HIGH_RISK"
    }
    """


@pytest.fixture
def feature_json_invalid_leakage():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": true,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": 4
    }
    """


@pytest.fixture
def modeling_feature_server_data(project_id):
    return {
        "featureType": "Numeric",
        "lowInformation": False,
        "uniqueCount": 252,
        "name": "a_feat",
        "min": 0.052,
        "importance": 0.0039430549403626225,
        "max": 0.15,
        "dateFormat": None,
        "median": 0.091,
        "stdDev": 0.014,
        "projectId": project_id,
        "naCount": 0,
        "parentFeatureNames": ["original"],
        "mean": 0.092,
    }


@pytest.fixture
def summarized_categorical_feature_json():
    return """
        {
            "id": 12,
            "name": "bids_device[count]",
            "featureType": "Summarized Categorical",
            "importance": 1,
            "lowInformation": false,
            "uniqueCount": 200,
            "naCount": 0,
            "dateFormat": null,
            "projectId": "556cdfbb100d2b0e88585195",
            "min": null,
            "max": null,
            "mean": null,
            "median": null,
            "stdDev": null,
            "timeSeriesEligible": false,
            "timeSeriesEligibilityReason": "notADate",
            "timeStep": null,
            "timeUnit": null,
            "targetLeakage": "SKIPPED_DETECTION",
            "keySummary": [
            {
                "key": "phone1",
                "summary": {
                    "min": 100,
                    "max": 200,
                    "median": 0,
                    "mean": 23.4,
                    "stdDev": 0,
                    "pctRows": 44.96
                }
            },
            {
                "key": "phone2",
                "summary": {
                    "min": 0,
                    "max": 4798,
                    "median": 0,
                    "mean": 24.91,
                    "stdDev": 194.917,
                    "pctRows": 38.75
                }
            }]
        }
        """


@pytest.fixture
def summarized_categorical_feature_json_empty_keys():
    return """
        {
            "id": 12,
            "name": "bids_device[count]",
            "featureType": "Summarized Categorical",
            "importance": 1,
            "lowInformation": false,
            "uniqueCount": 200,
            "naCount": 0,
            "dateFormat": null,
            "projectId": "556cdfbb100d2b0e88585195",
            "min": null,
            "max": null,
            "mean": null,
            "median": null,
            "stdDev": null,
            "timeSeriesEligible": false,
            "timeSeriesEligibilityReason": "notADate",
            "timeStep": null,
            "timeUnit": null,
            "targetLeakage": "SKIPPED_DETECTION",
            "keySummary": [
            {
                "key": " ",
                "summary": {
                    "min": 100,
                    "max": 200,
                    "median": 0,
                    "mean": 23.4,
                    "stdDev": 0,
                    "pctRows": 44.96
                }
            },
            {
                "key": "%",
                "summary": {
                    "min": 0,
                    "max": 4798,
                    "median": 0,
                    "mean": 24.91,
                    "stdDev": 194.917,
                    "pctRows": 38.75
                }
            }]
        }
        """


def get_feature_response_multicategorical(multilabel_insights_key, project_id, feature_name):
    return {
        u"dataQualities": u"NO_ISSUES_FOUND",
        u"dateFormat": None,
        u"featureLineageId": None,
        u"featureType": u"Multicategorical",
        u"id": 10,
        u"importance": 1,
        u"keySummary": [
            {
                u"key": u"interest_low",
                u"summary": {
                    u"max": 1,
                    u"mean": 0.7375,
                    u"median": 1.0,
                    u"min": 0,
                    u"pctRows": 73.75,
                    u"stdDev": 0.44137435443704354,
                },
            },
            {
                u"key": u"price_very_high",
                u"summary": {
                    u"max": 1,
                    u"mean": 0.225,
                    u"median": 0.0,
                    u"min": 0,
                    u"pctRows": 22.5,
                    u"stdDev": 0.41889342094762905,
                },
            },
        ],
        u"lowInformation": False,
        u"max": None,
        u"mean": None,
        u"median": None,
        u"min": None,
        u"multilabelInsights": {u"multilabelInsightsKey": multilabel_insights_key},
        u"naCount": 0,
        u"name": feature_name,
        u"parentFeatureNames": [],
        u"projectId": project_id,
        u"stdDev": None,
        u"targetLeakage": u"SKIPPED_DETECTION",
        u"targetLeakageReason": u"no target leakage was detected",
        u"timeSeriesEligibilityReason": u"notADate",
        u"timeSeriesEligible": False,
        u"timeStep": None,
        u"timeUnit": None,
        u"uniqueCount": 13,
    }


@pytest.fixture
def feature_lineage_json():
    return """
        {
            "steps": [
                {
                    "dataType": "Numeric",
                    "parents": [],
                    "id": 0,
                    "stepType": "generatedColumn",
                    "name": "transactions_lower[Description] (30 days entropy)"
                },
                {
                    "isTimeAware": false,
                    "name": "Entropy",
                    "stepType": "action",
                    "groupBy": [],
                    "timeInfo": null,
                    "parents": [
                        0
                    ],
                    "arguments": {},
                    "actionType": "safer",
                    "id": 1,
                    "description": "Calculates the entropy over distribution of the categorie."
                },
                {
                    "joinInfo": {
                        "leftTable": {
                            "datasteps": [
                                4
                            ],
                            "columns": [
                                "CustomerID"
                            ]
                        },
                        "rightTable": {
                            "datasteps": [
                                5
                            ],
                            "columns": [
                                "customer_id"
                            ]
                        },
                        "joinType": "left"
                    },
                    "isTimeAware": false,
                    "id": 2,
                    "stepType": "join",
                    "parents": [
                        1
                    ]
                },
                {
                    "isTimeAware": true,
                    "name": "Value counts",
                    "stepType": "action",
                    "groupBy": [
                        "customer_id"
                    ],
                    "timeInfo": {
                        "duration": {
                            "duration": 30,
                            "timeUnit": "DAY"
                        },
                        "latest": {
                            "duration": 0,
                            "timeUnit": "DAY"
                        }
                    },
                    "parents": [
                        2
                    ],
                    "arguments": {
                        "fdwMeta": {
                            "duration": 30,
                            "timeUnit": "DAY"
                        },
                        "enableHashing": false
                    },
                    "actionType": "safer",
                    "id": 3,
                    "description": \
"Counts values of a given type within the feature derivation window.\\n\
    The whole dataset is in use when time aware modeling is disabled."
                },
                {
                    "name": "Primary dataset",
                    "catalogVersionId": null,
                    "stepType": "data",
                    "catalogId": null,
                    "parents": [
                        2
                    ],
                    "id": 4,
                    "columns": [
                        {
                            "dataType": "Numeric",
                            "isInput": false,
                            "name": "date",
                            "isCutoff": true
                        },
                        {
                            "dataType": "Categorical",
                            "isInput": false,
                            "name": "CustomerID",
                            "isCutoff": false
                        }
                    ]
                },
                {
                    "name": "transactions_lower",
                    "catalogVersionId": "5dcab3fe3a35220406f281e4",
                    "stepType": "data",
                    "catalogId": "5dcab3fe3a35220406f281e3",
                    "parents": [
                        3
                    ],
                    "id": 5,
                    "columns": [
                        {
                            "dataType": "Categorical",
                            "isInput": true,
                            "name": "Description",
                            "isCutoff": false
                        },
                        {
                            "dataType": "Numeric",
                            "isInput": false,
                            "name": "Date",
                            "isCutoff": true
                        },
                        {
                            "dataType": "Categorical",
                            "isInput": false,
                            "name": "customer_id",
                            "isCutoff": false
                        }
                    ]
                }
            ]
        }
        """


@pytest.fixture
def modeling_features_list_server_data(modeling_feature_server_data):
    return {"count": 1, "next": None, "previous": None, "data": [modeling_feature_server_data]}


@pytest.fixture
def modeling_features_with_next_page_server_data(modeling_feature_server_data, project_url):
    next_page_url = "{}modelingFeatures/?offset=1&limit=1".format(project_url)
    feat = dict(modeling_feature_server_data)
    feat["name"] = "first_page"
    return {"count": 1, "next": next_page_url, "previous": None, "data": [feat]}


@pytest.fixture
def modeling_features_with_previous_page_server_data(modeling_feature_server_data, project_url):
    previous_page_url = "{}modelingFeatures/?offset=0&limit=1".format(project_url)
    feat = dict(modeling_feature_server_data)
    feat["name"] = "second_page"
    return {"count": 1, "next": None, "previous": previous_page_url, "data": [feat]}


@pytest.fixture
def feature_sans_importance_json():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": null,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "556cdfbb100d2b0e88585195",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def claim_amount_json():
    return """
    {
        "id": 34,
        "name": "Claim Amount",
        "featureType": "Numeric",
        "importance": null,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "556cdfbb100d2b0e88585195",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def low_info_feature_json():
    """The information for a feature that has been marked as low info

    DataRobot does not finish gathering the naCount or generate a featureType
    in these situations
    """
    return """
    {
    "featureType": null,
    "lowInformation": true,
    "name": "mths_since_last_major_derog",
    "uniqueCount": 0,
    "importance": null,
    "id": 25,
    "naCount": null,
    "dateFormat": null,
    "projectId": "556cdfbb100d2b0e88585195",
    "min": null,
    "max": null,
    "mean": null,
    "median": null,
    "stdDev": null,
    "timeSeriesEligible": false,
    "timeSeriesEligibilityReason": "notADate",
    "timeStep": null,
    "timeUnit": null,
    "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def float_feature_json():
    return """
    {
        "uniqueCount": 1,
        "lowInformation": false,
        "name": "float",
        "dateFormat": null,
        "importance": 0.2648777523946366,
        "projectId": "556cdfbb100d2b0e88585195",
        "featureType": "Date",
        "id": 10,
        "naCount": 0,
        "min": 123.45,
        "max": 123.45,
        "mean": 123.45,
        "median": 123.45,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def date_feature_json():
    return """
    {
        "uniqueCount": 157,
        "lowInformation": false,
        "name": "dates",
        "dateFormat": "%Y-%m-%d %H:%M:%S",
        "importance": 0.2648777523946366,
        "projectId": "556cdfbb100d2b0e88585195",
        "featureType": "Date",
        "id": 1,
        "naCount": 0,
        "min": "2017-03-04T00:00:00",
        "max": "2017-06-15T04:40:39",
        "mean": "2017-04-24T14:20:19",
        "median": "2017-04-24T14:20:19",
        "stdDev": "30.089 days",
        "timeSeriesEligible": true,
        "timeSeriesEligibilityReason": "suitable",
        "timeStep": 3,
        "timeUnit": "MONTH",
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def histogram_plot_json():
    return """
    {
        "plot":
        [
            {"label": "5.0", "count": 50.5, "target": 3.01},
            {"label": "15.1", "count": 150.6, "target": 2.9529411764},
            {"label": "25.2", "count": 250.7, "target": 2.8980769230},
            {"label": "==Missing==", "count": 350.8, "target": 2.8452830188},
            {"label": "=All Other=", "count": 450.9, "target": 2.79444444446}
        ]
    }
    """


@pytest.fixture
def histogram_plot_without_target_json():
    return """
    {
        "plot":
        [
            {"label": "5.0", "count": 50, "target": null},
            {"label": "15.1", "count": 150, "target": null},
            {"label": "==Missing==", "count": 350, "target": null},
            {"label": "=All Other=", "count": 450, "target": null}
        ]
    }
    """


@pytest.fixture
def histogram_plot_for_dataset_missing_target_field():
    return """
    {
        "plot":
        [
            {"label": "5.0", "count": 50},
            {"label": "15.1", "count": 150},
            {"label": "==Missing==", "count": 350},
            {"label": "=All Other=", "count": 450}
        ]
    }
    """


@pytest.fixture
def feature_server_data(feature_json):
    return json.loads(feature_json)


def test_future_proof(feature_server_data):
    Feature.from_server_data(dict(feature_server_data, future="key"))


@responses.activate
def test_features(list_features_json):
    responses.add(
        responses.GET, "https://host_name.com/projects/p-id/features/", body=list_features_json
    )
    feature = Project("p-id").get_features()[0]
    assert isinstance(feature, Feature)
    assert feature.id == 34
    assert feature.name == "Claim_Amount"
    assert feature.feature_type == "Numeric"
    assert feature.importance == 1
    assert feature.low_information is False
    assert feature.unique_count == 200
    assert feature.na_count == 0
    assert feature.project_id == "556cdfbb100d2b0e88585195"
    assert feature.min == 1
    assert feature.max == 1
    assert feature.median == 1
    assert feature.median == 1
    assert feature.std_dev == 0
    assert feature.time_series_eligible is False
    assert feature.time_series_eligibility_reason == "notADate"
    assert feature.time_step is None
    assert feature.time_unit is None


@responses.activate
def test_feature(feature_sans_importance_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/f-id/",
        body=feature_sans_importance_json,
    )
    feature = Feature.get("p-id", "f-id")
    assert isinstance(feature, Feature)
    assert feature.id == 34
    assert feature.name == "Claim_Amount"
    assert feature.feature_type == "Numeric"
    assert feature.importance is None
    assert feature.feature_lineage_id is None
    assert feature.low_information is False
    assert feature.unique_count == 200
    assert feature.na_count == 0
    assert feature.date_format is None
    assert feature.project_id == "556cdfbb100d2b0e88585195"
    assert feature.min == 1
    assert feature.max == 1
    assert feature.median == 1
    assert feature.median == 1
    assert feature.std_dev == 0
    assert feature.target_leakage == "SKIPPED_DETECTION"


@responses.activate
def test_feature_no_leakage(feature_json_no_leakage):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/f-id/",
        body=feature_json_no_leakage,
    )
    feature = Feature.get("p-id", "f-id")
    assert isinstance(feature, Feature)
    assert feature.low_information is False
    assert feature.target_leakage == "FALSE"


@responses.activate
def test_feature_leakage(feature_json_leakage):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/f-id/",
        body=feature_json_leakage,
    )
    feature = Feature.get("p-id", "f-id")
    assert isinstance(feature, Feature)
    assert feature.low_information is True
    assert feature.target_leakage == "HIGH_RISK"


@responses.activate
def test_feature_invalid_leakage(feature_json_invalid_leakage):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/f-id/",
        body=feature_json_invalid_leakage,
    )
    with pytest.raises(DataError):
        Feature.get("p-id", "f-id")


@responses.activate
def test_feature_with_low_info(low_info_feature_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/f-id/",
        body=low_info_feature_json,
    )
    feature = Feature.get("p-id", "f-id")
    assert feature.na_count is None
    assert feature.feature_type is None
    assert feature.min is None
    assert feature.max is None
    assert feature.median is None
    assert feature.median is None
    assert feature.std_dev is None


@responses.activate
def test_feature_with_space(claim_amount_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/Claim%20Amount/",
        body=claim_amount_json,
    )
    feature = Feature.get("p-id", "Claim Amount")
    assert isinstance(feature, Feature)


@responses.activate
def test_float_feature(float_feature_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/float/",
        body=float_feature_json,
    )
    feature = Feature.get("p-id", "float")
    assert isinstance(feature, Feature)
    assert feature.min == 123.45
    assert feature.max == 123.45
    assert feature.mean == 123.45
    assert feature.median == 123.45
    assert feature.std_dev == 0


@responses.activate
def test_date_feature(date_feature_json):
    responses.add(
        responses.GET, "https://host_name.com/projects/p-id/features/dates/", body=date_feature_json
    )
    feature = Feature.get("p-id", "dates")
    assert isinstance(feature, Feature)
    assert feature.date_format == json.loads(date_feature_json)["dateFormat"]
    assert feature.min == "2017-03-04T00:00:00"
    assert feature.max == "2017-06-15T04:40:39"
    assert feature.mean == "2017-04-24T14:20:19"
    assert feature.median == "2017-04-24T14:20:19"
    assert feature.std_dev == "30.089 days"
    assert feature.time_series_eligible is True
    assert feature.time_series_eligibility_reason == "suitable"
    assert feature.time_step == 3
    assert feature.time_unit == "MONTH"


@responses.activate
def test_summarized_categorical_feature(summarized_categorical_feature_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/bids_device/",
        body=summarized_categorical_feature_json,
    )
    feature = Feature.get("p-id", "bids_device")
    assert isinstance(feature, Feature)
    assert feature.name == "bids_device[count]"
    assert feature.min is None
    assert feature.max is None
    assert feature.mean is None
    assert feature.median is None
    assert feature.std_dev is None
    assert feature.key_summary
    assert len(feature.key_summary) == 2
    assert feature.key_summary[0]["key"] == "phone1"
    assert len(feature.key_summary[0]["summary"]) == 6
    assert feature.key_summary[1]["key"] == "phone2"
    assert len(feature.key_summary[0]["summary"]) == 6


@responses.activate
def test_summarized_categorical_feature_empty_key(summarized_categorical_feature_json_empty_keys):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/bids_device/",
        body=summarized_categorical_feature_json_empty_keys,
    )
    feature = Feature.get("p-id", "bids_device")
    assert isinstance(feature, Feature)
    assert feature.name == "bids_device[count]"
    assert feature.min is None
    assert feature.max is None
    assert feature.mean is None
    assert feature.median is None
    assert feature.std_dev is None
    assert feature.key_summary
    assert len(feature.key_summary) == 2
    assert feature.key_summary[0]["key"] == " "
    assert len(feature.key_summary[0]["summary"]) == 6
    assert feature.key_summary[1]["key"] == "%"
    assert len(feature.key_summary[0]["summary"]) == 6


@responses.activate
def test_multicategorical_feature():
    multilabel_insights_key = "foo_insights_key"
    project_id = "p-id"
    feature_name = "feature_name"
    response = get_feature_response_multicategorical(
        multilabel_insights_key, project_id, feature_name
    )
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/bids_device/",
        body=json.dumps(response),
    )
    feature = Feature.get("p-id", "bids_device")
    assert feature.multilabel_insights_key == multilabel_insights_key


@responses.activate
def test_multicategorical_feature_no_multilabel_insights_in_response():
    multilabel_insights_key = "foo_insights_key"
    project_id = "p-id"
    feature_name = "feature_name"
    response = get_feature_response_multicategorical(
        multilabel_insights_key, project_id, feature_name
    )
    del response["multilabelInsights"]
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/features/bids_device/",
        body=json.dumps(response),
    )
    feature = Feature.get("p-id", "bids_device")
    assert feature.multilabel_insights_key is None


@responses.activate
def test_get_multiseries_properties_pre_computed(
    date_feature_json, multiseries_precomputed_properties
):
    feature = Feature.from_server_data(json.loads(date_feature_json))

    properties = feature.get_multiseries_properties(["series_id"])
    assert properties == {
        "time_series_eligible": True,
        "time_step": multiseries_precomputed_properties["timeStep"],
        "time_unit": multiseries_precomputed_properties["timeUnit"],
    }
    assert len(responses.calls) == 1


@responses.activate
def test_get_multiseries_properties_with_computation(
    date_feature_json, multiseries_postcomputed_properties
):
    feature = Feature.from_server_data(json.loads(date_feature_json))

    properties = feature.get_multiseries_properties(["series_id"])
    assert properties == {
        "time_series_eligible": True,
        "time_step": multiseries_postcomputed_properties["timeStep"],
        "time_unit": multiseries_postcomputed_properties["timeUnit"],
    }
    assert len(responses.calls) == 4


@responses.activate
@pytest.mark.usefixtures("multiseries_ineligible_properties")
def test_get_multiseries_properties_not_eligible(float_feature_json):
    feature = Feature.from_server_data(json.loads(float_feature_json))

    properties = feature.get_multiseries_properties(["series_id"])
    assert properties == {"time_series_eligible": False, "time_step": None, "time_unit": None}
    assert len(responses.calls) == 4


def test_feature_with_non_ascii_name(feature_server_data, unicode_string):
    data_copy = dict(feature_server_data)
    data_copy["name"] = unicode_string
    feature = Feature.from_server_data(data_copy)
    print(feature)  # actually part of the test - this used to fail (testing __repr__)


@pytest.mark.usefixtures("known_warning")
def test_bc_instantiate_feature_from_dict(feature_server_data):
    feature = Feature(feature_server_data)
    assert feature.id == feature_server_data["id"]
    assert feature.name == feature_server_data["name"]
    assert feature.feature_type == feature_server_data["featureType"]
    assert feature.importance == feature_server_data["importance"]
    assert feature.low_information == feature_server_data["lowInformation"]
    assert feature.unique_count == feature_server_data["uniqueCount"]
    assert feature.na_count == feature_server_data["naCount"]
    assert feature.project_id == "556cdfbb100d2b0e88585195"


def test_modeling_feature_future_proof(modeling_feature_server_data):
    future_data = dict(modeling_feature_server_data, new="new")
    ModelingFeature.from_server_data(future_data)


@responses.activate
def test_get_modeling_feature(modeling_feature_server_data, project_url, project_id):
    name = modeling_feature_server_data["name"]
    responses.add(
        responses.GET,
        "{}modelingFeatures/a_feat/".format(project_url),
        json=modeling_feature_server_data,
    )

    feature = ModelingFeature.get(project_id, name)
    assert feature.project_id == project_id
    assert feature.name == modeling_feature_server_data["name"]
    assert feature.feature_type == modeling_feature_server_data["featureType"]
    assert feature.importance == modeling_feature_server_data["importance"]
    assert feature.low_information == modeling_feature_server_data["lowInformation"]
    assert feature.unique_count == modeling_feature_server_data["uniqueCount"]
    assert feature.na_count == modeling_feature_server_data["naCount"]
    assert feature.date_format == modeling_feature_server_data["dateFormat"]
    assert feature.min == modeling_feature_server_data["min"]
    assert feature.max == modeling_feature_server_data["max"]
    assert feature.mean == modeling_feature_server_data["mean"]
    assert feature.median == modeling_feature_server_data["median"]
    assert feature.std_dev == modeling_feature_server_data["stdDev"]
    assert feature.parent_feature_names == modeling_feature_server_data["parentFeatureNames"]


@responses.activate
def test_list_modeling_features(modeling_features_list_server_data, project_url, project):
    responses.add(
        responses.GET,
        "{}modelingFeatures/".format(project_url),
        json=modeling_features_list_server_data,
    )
    feats = project.get_modeling_features()

    assert len(feats) == len(modeling_features_list_server_data["data"])
    assert isinstance(feats[0], ModelingFeature)
    assert feats[0].name == modeling_features_list_server_data["data"][0]["name"]


@responses.activate
def test_list_modeling_features_paginated(
    modeling_features_with_next_page_server_data,
    modeling_features_with_previous_page_server_data,
    project,
    project_url,
):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}modelingFeatures/".format(project_url),
            json=modeling_features_with_next_page_server_data,
        )
        rsps.add(
            responses.GET,
            "{}modelingFeatures/".format(project_url),
            json=modeling_features_with_previous_page_server_data,
        )
        feats = project.get_modeling_features(batch_size=1)

        first_page_req = rsps.calls[0].request
        second_page_req = rsps.calls[1].request
        assert {"limit": ["1"]} == parse_qs(urlparse(first_page_req.url).query)
        assert {"limit": ["1"], "offset": ["1"]} == parse_qs(urlparse(second_page_req.url).query)

    assert len(feats) == 2
    names = {"first_page", "second_page"}
    assert {feat.name for feat in feats} == names


@responses.activate
@pytest.mark.parametrize("histogram_class", [FeatureHistogram, DatasetFeatureHistogram])
def test_feature_histogram(
    histogram_class, project_url, project_id, modeling_feature_server_data, histogram_plot_json
):
    feature_name = modeling_feature_server_data["name"]
    project_url = _convert_project_url_for_dataset_histogram(histogram_class, project_url)

    responses.add(
        responses.GET,
        "{}featureHistograms/{}/".format(project_url, feature_name),
        body=histogram_plot_json,
    )

    histogram = histogram_class.get(project_id, feature_name)

    assert len(histogram.plot) == 5
    assert histogram.plot[0]["label"] == "5.0"
    assert histogram.plot[0]["count"] == 50.5
    assert histogram.plot[0]["target"] == 3.01


@responses.activate
def test_dataset_feature_histogram_with_missing_target_field(
    project_url, project_id, histogram_plot_for_dataset_missing_target_field
):
    feature_name = "joe"
    dataset_url = project_url.replace("projects", "datasets")
    responses.add(
        responses.GET,
        "{}featureHistograms/{}/".format(dataset_url, feature_name),
        body=histogram_plot_for_dataset_missing_target_field,
    )

    histogram = DatasetFeatureHistogram.get(project_id, feature_name)

    assert len(histogram.plot) == 4
    assert histogram.plot[0]["label"] == "5.0"
    assert histogram.plot[0]["count"] == 50
    assert histogram.plot[0]["target"] is None


@responses.activate
@pytest.mark.parametrize("histogram_class", [FeatureHistogram, DatasetFeatureHistogram])
def test_feature_histogram_without_target(
    histogram_class,
    project_url,
    project_id,
    modeling_feature_server_data,
    histogram_plot_without_target_json,
):
    feature_name = modeling_feature_server_data["name"]
    project_url = _convert_project_url_for_dataset_histogram(histogram_class, project_url)
    responses.add(
        responses.GET,
        "{}featureHistograms/{}/".format(project_url, feature_name),
        body=histogram_plot_without_target_json,
    )

    histogram = histogram_class.get(project_id, feature_name)

    assert len(histogram.plot) == 4
    assert histogram.plot[0]["label"] == "5.0"
    assert histogram.plot[0]["count"] == 50
    assert histogram.plot[0]["target"] is None


@responses.activate
@pytest.mark.parametrize("histogram_class", [FeatureHistogram, DatasetFeatureHistogram])
def test_feature_histogram_with_bin_limit(
    histogram_class, project_url, project_id, modeling_feature_server_data, histogram_plot_json
):
    bin_limit = 5
    feature_name = modeling_feature_server_data["name"]
    project_url = _convert_project_url_for_dataset_histogram(histogram_class, project_url)
    responses.add(
        responses.GET,
        "{}featureHistograms/{}/?binLimit={}".format(project_url, feature_name, bin_limit),
        body=histogram_plot_json,
    )

    histogram = histogram_class.get(project_id, feature_name, bin_limit)

    assert len(histogram.plot) == 5
    assert histogram.plot[0]["label"] == "5.0"
    assert histogram.plot[0]["count"] == 50.5
    assert histogram.plot[0]["target"] == 3.01


@responses.activate
@pytest.mark.parametrize("histogram_class", [FeatureHistogram, DatasetFeatureHistogram])
def test_feature_histogram_using_summarized_categorical_key(
    histogram_class, project_url, project_id, histogram_plot_json
):
    key_name = "sc_key1"
    feature_name = "bids_device"
    project_url = _convert_project_url_for_dataset_histogram(histogram_class, project_url)
    responses.add(
        responses.GET,
        "{}featureHistograms/{}/?key={}".format(project_url, feature_name, key_name),
        body=histogram_plot_json,
    )

    histogram = histogram_class.get(project_id, feature_name=feature_name, key_name=key_name)

    assert len(histogram.plot) == 5
    assert histogram.plot[0]["label"] == "5.0"
    assert histogram.plot[0]["count"] == 50.5
    assert histogram.plot[0]["target"] == 3.01


def _convert_project_url_for_dataset_histogram(feature_class, project_url):
    if issubclass(feature_class, DatasetFeatureHistogram):
        project_url = project_url.replace("projects", "datasets")
    return project_url


@responses.activate
def test_feature_histogram_from_feature(
    project_url, project_id, list_features_json, histogram_plot_json
):
    responses.add(
        responses.GET, "https://host_name.com/projects/p-id/features/", body=list_features_json
    )
    feature = Project("p-id").get_features()[0]

    bin_limit = 3
    responses.add(
        responses.GET,
        "{}featureHistograms/{}/?binLimit={}".format(project_url, feature.name, bin_limit),
        body=histogram_plot_json,
    )

    actualPlot = feature.get_histogram(bin_limit).plot
    expectedPlot = FeatureHistogram.get(project_id, feature.name, bin_limit).plot
    assert actualPlot == expectedPlot


@responses.activate
def test_feature_histogram_from_modeling_feature(
    project_url, project_id, modeling_feature_server_data, histogram_plot_json
):
    name = modeling_feature_server_data["name"]
    responses.add(
        responses.GET,
        "{}modelingFeatures/{}/".format(project_url, name),
        json=modeling_feature_server_data,
    )
    feature = ModelingFeature.get(project_id, name)

    bin_limit = 3
    responses.add(
        responses.GET,
        "{}featureHistograms/{}/?binLimit={}".format(project_url, feature.name, bin_limit),
        body=histogram_plot_json,
    )

    actualPlot = feature.get_histogram(bin_limit).plot
    expectedPlot = FeatureHistogram.get(project_id, feature.name, bin_limit).plot
    assert actualPlot == expectedPlot


def assert_dataset_feature(feature_obj, feature_json):
    # type: (DatasetFeature, dict) -> None
    keys = [
        "id",
        "dataset_id",
        "dataset_version_id",
        "name",
        "feature_type",
        "low_information",
        "unique_count",
        "na_count",
        "date_format",
        "min",
        "max",
        "mean",
        "median",
        "std_dev",
        "time_series_eligible",
        "time_series_eligibility_reason",
        "time_step",
        "time_unit",
        "target_leakage",
        "target_leakage_reason",
    ]
    assert isinstance(feature_obj, DatasetFeature)
    for key in keys:
        assert getattr(feature_obj, key) == feature_json.get(camelize(key))


@responses.activate
def test_get_interaction_feature(interaction_feature_server_data, project_url, project_id):

    responses.add(
        responses.GET,
        "{}interactionFeatures/int0/".format(project_url),
        json=interaction_feature_server_data,
    )

    feature = InteractionFeature.get(project_id, "int0")
    assert feature.rows == interaction_feature_server_data["rows"]
    assert feature.bars == interaction_feature_server_data["bars"]
    assert feature.bubbles == interaction_feature_server_data["bubbles"]
    assert set(feature.source_columns) == set(interaction_feature_server_data["source_columns"])


@responses.activate
def test_get_feature_lineage(project_id, project_url, feature_lineage_json):
    feature_lineage_id = "5f05b824949de873c7300a4c"

    responses.add(
        responses.GET,
        "{}featureLineages/{}/".format(project_url, feature_lineage_id),
        body=feature_lineage_json,
    )

    feature_lineage = FeatureLineage.get(project_id, feature_lineage_id)
    assert len(feature_lineage.steps) == 6

    assert [x["step_type"] for x in feature_lineage.steps] == [
        "generatedColumn",
        "action",
        "join",
        "action",
        "data",
        "data",
    ]
    assert [x["parents"] for x in feature_lineage.steps] == [[], [0], [1], [2], [2], [3]]
    assert feature_lineage.steps[3]["is_time_aware"]
    assert feature_lineage.steps[3]["time_info"] == {
        "duration": {"duration": 30, "time_unit": "DAY"},
        "latest": {"duration": 0, "time_unit": "DAY"},
    }
    assert "arguments" not in feature_lineage.steps[3]
    assert "classname" not in feature_lineage.steps[3]
    assert "action_type" not in feature_lineage.steps[3]
    assert "actionType" not in feature_lineage.steps[3]


class TestDatasetFeature(object):
    @staticmethod
    def convert_to_dataset_feature(feature_json):
        dataset_id = str(uuid4())
        version_id = str(uuid4())
        feature_json["datasetId"] = dataset_id
        feature_json["datasetVersionId"] = version_id
        feature_json["targetLeakageReason"] = "ummmm. Do not question target leakage!"
        return feature_json

    def test_instantiation(
        self,
        feature_json,
        feature_json_leakage,
        feature_json_no_leakage,
        feature_sans_importance_json,
        date_feature_json,
        low_info_feature_json,
        claim_amount_json,
        float_feature_json,
        summarized_categorical_feature_json,
        summarized_categorical_feature_json_empty_keys,
    ):

        for feature_string in [
            feature_json,
            feature_json_leakage,
            feature_json_no_leakage,
            feature_sans_importance_json,
            date_feature_json,
            low_info_feature_json,
            claim_amount_json,
            float_feature_json,
            summarized_categorical_feature_json,
            summarized_categorical_feature_json_empty_keys,
        ]:
            feature_json = json.loads(feature_string)
            new_json = self.convert_to_dataset_feature(feature_json)
            new_obj = DatasetFeature.from_server_data(new_json)
            assert_dataset_feature(new_obj, new_json)

    @responses.activate
    def test_get_histogram(self, feature_json, unittest_endpoint, histogram_plot_json):
        dataset_feature_json = self.convert_to_dataset_feature(json.loads(feature_json))
        feature = DatasetFeature.from_server_data(dataset_feature_json)

        url = "{}/datasets/{}/featureHistograms/{}/".format(
            unittest_endpoint, feature.dataset_id, feature.name
        )
        responses.add(responses.GET, url, body=histogram_plot_json)

        result = feature.get_histogram()
        assert result.plot == json.loads(histogram_plot_json)["plot"]

        request = responses.calls[0].request
        url_tail = "{}/featureHistograms/{}/".format(feature.dataset_id, feature.name)
        assert request.url.endswith(url_tail)

    @responses.activate
    def test_get_histogram_with_bin(self, feature_json, unittest_endpoint, histogram_plot_json):
        dataset_feature_json = self.convert_to_dataset_feature(json.loads(feature_json))
        feature = DatasetFeature.from_server_data(dataset_feature_json)

        url = "{}/datasets/{}/featureHistograms/{}/?binLimit=5".format(
            unittest_endpoint, feature.dataset_id, feature.name
        )
        responses.add(responses.GET, url, body=histogram_plot_json)

        result = feature.get_histogram(bin_limit=5)
        assert result.plot == json.loads(histogram_plot_json)["plot"]

        request = responses.calls[0].request
        url_tail = "{}/featureHistograms/{}/?binLimit=5".format(feature.dataset_id, feature.name)
        assert request.url.endswith(url_tail)


class TestMulticategoricalSpecificMethods(object):
    @pytest.fixture
    def multilabel_insights_key(self):
        return "edca06201e9ba259981661145de283fd"

    @pytest.fixture
    def project_id(self):
        return "601bddf788dfc14dddd926d0"

    @pytest.fixture
    def feature_name(self):
        return "target"

    @pytest.fixture
    def feature_response_multicategorical(self, multilabel_insights_key, project_id, feature_name):
        return get_feature_response_multicategorical(
            multilabel_insights_key, project_id, feature_name
        )

    @pytest.fixture
    def multilabel_insights_histogram_response(self, project_id, feature_name):
        response = {
            u"featureName": feature_name,
            u"histogram": [
                {
                    u"label": u"interest_low",
                    u"plot": [
                        {u"labelRelevance": 0, u"rowCount": 42, u"rowPct": 26.25},
                        {u"labelRelevance": 1, u"rowCount": 118, u"rowPct": 73.75},
                    ],
                },
                {
                    u"label": u"price_very_high",
                    u"plot": [
                        {u"labelRelevance": 0, u"rowCount": 124, u"rowPct": 77.5},
                        {u"labelRelevance": 1, u"rowCount": 36, u"rowPct": 22.5},
                    ],
                },
            ],
            u"projectId": project_id,
        }
        return response

    @pytest.fixture
    def expected_histogram_values(self):
        return [
            {
                u"label": u"interest_low",
                u"plot": [
                    {u"label_relevance": 0, u"row_count": 42, u"row_pct": 26.25},
                    {u"label_relevance": 1, u"row_count": 118, u"row_pct": 73.75},
                ],
            },
            {
                u"label": u"price_very_high",
                u"plot": [
                    {u"label_relevance": 0, u"row_count": 124, u"row_pct": 77.5},
                    {u"label_relevance": 1, u"row_count": 36, u"row_pct": 22.5},
                ],
            },
        ]

    @responses.activate
    def test_multicategorical_histogram_get(
        self,
        multilabel_insights_key,
        multilabel_insights_histogram_response,
        project_id,
        feature_name,
        expected_histogram_values,
    ):
        responses.add(
            responses.GET,
            "https://host_name.com/multilabelInsights/{multilabel_insights_key}/histogram/".format(
                multilabel_insights_key=multilabel_insights_key
            ),
            body=json.dumps(multilabel_insights_histogram_response),
        )
        histogram = MulticategoricalHistogram.get(multilabel_insights_key)

        self._assert_histogram(histogram, feature_name, expected_histogram_values)

    def _assert_histogram(self, histogram, feature_name, expected_histogram_values):
        assert histogram.values == expected_histogram_values
        assert histogram.feature_name == feature_name

    def test_multicategorical_histogram_to_dataframe(self, expected_histogram_values, feature_name):
        multicategorical_histogram = MulticategoricalHistogram(
            feature_name=feature_name, histogram=expected_histogram_values
        )
        actual = multicategorical_histogram.to_dataframe()

        expected = pd.DataFrame(
            [
                {
                    "feature_name": feature_name,
                    "label": "interest_low",
                    "label_relevance": 0,
                    "row_count": 42,
                    "row_pct": 26.25,
                },
                {
                    "feature_name": feature_name,
                    "label": "interest_low",
                    "label_relevance": 1,
                    "row_count": 118,
                    "row_pct": 73.75,
                },
                {
                    "feature_name": feature_name,
                    "label": "price_very_high",
                    "label_relevance": 0,
                    "row_count": 124,
                    "row_pct": 77.5,
                },
                {
                    "feature_name": feature_name,
                    "label": "price_very_high",
                    "label_relevance": 1,
                    "row_count": 36,
                    "row_pct": 22.5,
                },
            ]
        )

        pd.testing.assert_frame_equal(actual, expected)

    def test_feature_get_multicategorical_histogram_not_a_multicategorical_feature(self):
        feature = Feature(id=1324, feature_type="Regression")
        error_message = (
            "Multicategorical Histograms are only available for features"
            "with feature_type Multicategorical. "
            "This feature is of type: Regression"
        )
        with pytest.raises(InvalidUsageError, match=error_message):
            feature.get_multicategorical_histogram()

    def test_feature_get_multicategorical_histogram_multicategorical_feature_but_no_insights_key(
        self,
    ):
        feature = Feature(id=1324, feature_type="Multicategorical", multilabel_insights=None)
        error_message = "A valid multilabel_insights_key is required, but not present"
        with pytest.raises(ValueError, match=error_message):
            feature.get_multicategorical_histogram()

    @responses.activate
    def test_feature_get_multicategorical_histogram_via_feature(
        self,
        feature_response_multicategorical,
        multilabel_insights_key,
        multilabel_insights_histogram_response,
        expected_histogram_values,
        project_id,
        feature_name,
    ):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/features/{}/".format(project_id, feature_name),
            json=feature_response_multicategorical,
        )
        responses.add(
            responses.GET,
            "https://host_name.com/multilabelInsights/{multilabel_insights_key}/histogram/".format(
                multilabel_insights_key=multilabel_insights_key
            ),
            json=multilabel_insights_histogram_response,
        )
        feature = Feature.get(project_id, feature_name)
        histogram = feature.get_multicategorical_histogram()

        self._assert_histogram(histogram, feature_name, expected_histogram_values)

    @responses.activate
    @pytest.mark.parametrize(
        ["getter_method", "patch_target"],
        [
            pytest.param(
                "get_pairwise_correlations", "datarobot.models.feature.PairwiseCorrelations"
            ),
            pytest.param(
                "get_pairwise_joint_probabilities",
                "datarobot.models.feature.PairwiseJointProbabilities",
            ),
            pytest.param(
                "get_pairwise_conditional_probabilities",
                "datarobot.models.feature.PairwiseConditionalProbabilities",
            ),
        ],
    )
    def test_get_pairwise_statistics_getter(
        self,
        getter_method,
        patch_target,
        feature_response_multicategorical,
        project_id,
        feature_name,
    ):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/features/{}/".format(project_id, feature_name),
            json=feature_response_multicategorical,
        )
        feature = Feature.get(project_id, feature_name)
        with mock.patch(patch_target, autospec=True) as p:
            returned = feature.__getattribute__(getter_method)()
        assert returned == p.get.return_value
        p.get.assert_called_once_with(feature.multilabel_insights_key)

    @pytest.mark.parametrize(
        "getter_method",
        [
            "get_pairwise_correlations",
            "get_pairwise_joint_probabilities",
            "get_pairwise_conditional_probabilities",
        ],
    )
    def test_pairwise_statistics_getter_multicategorical_feature_but_no_insights_key(
        self, getter_method
    ):
        feature = Feature(id=1324, feature_type="Multicategorical", multilabel_insights=None)
        error_message = "A valid multilabel_insights_key is required, but not present"
        with pytest.raises(ValueError, match=error_message):
            feature.__getattribute__(getter_method)()

    @pytest.mark.parametrize(
        ["getter_method", "expected_error_message"],
        [
            pytest.param(
                "get_pairwise_correlations",
                (
                    "Pairwise Correlation is only available for features"
                    "with feature_type Multicategorical. "
                    "This feature is of type: Regression"
                ),
                id="correlation",
            ),
            pytest.param(
                "get_pairwise_joint_probabilities",
                (
                    "Pairwise Joint Probability is only available for features"
                    "with feature_type Multicategorical. "
                    "This feature is of type: Regression"
                ),
                id="joint_probabilities",
            ),
            pytest.param(
                "get_pairwise_conditional_probabilities",
                (
                    "Pairwise Conditional Probability is only available for features"
                    "with feature_type Multicategorical. "
                    "This feature is of type: Regression"
                ),
                id="conditional_probabilities",
            ),
        ],
    )
    def test_pairwise_statistics_getter_not_a_multicategorical_feature(
        self, getter_method, expected_error_message
    ):
        feature = Feature(id=1324, feature_type="Regression")
        with pytest.raises(InvalidUsageError, match=expected_error_message):
            feature.__getattribute__(getter_method)()



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from copy import deepcopy
import json

import pytest
import responses
from six.moves.urllib.parse import urljoin

from datarobot import enums
from datarobot.enums import CUSTOM_MODEL_TARGET_TYPE
from tests.model.utils import change_response_value


@pytest.fixture
def model_url(project_url, model_id):
    return urljoin(project_url, "models/{}/".format(model_id))


@pytest.fixture
def feature_impact_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/featureImpact/".format(project_id, model_id)


@pytest.fixture
def multiclass_feature_impact_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/multiclassFeatureImpact/".format(
        project_id, model_id
    )


@pytest.fixture
def feature_impact_job_creation_response(feature_impact_url, job_url):
    responses.add(
        responses.POST,
        feature_impact_url,
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def multiclass_feature_impact_server_data():
    return {
        "ranRedundancyDetection": False,
        "classFeatureImpacts": [
            {
                "featureImpacts": [
                    {
                        "featureName": "Petal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.157,
                        "impactNormalized": 0.701,
                    },
                    {
                        "featureName": "Petal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.121,
                        "impactNormalized": 0.539,
                    },
                    {
                        "featureName": "Sepal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.225,
                        "impactNormalized": 1.0,
                    },
                    {
                        "featureName": "Sepal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.057,
                        "impactNormalized": 0.257,
                    },
                ],
                "class": "setosa",
            },
            {
                "featureImpacts": [
                    {
                        "featureName": "Petal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.302,
                        "impactNormalized": 1.0,
                    },
                    {
                        "featureName": "Petal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.229,
                        "impactNormalized": 0.758,
                    },
                    {
                        "featureName": "Sepal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.221,
                        "impactNormalized": 0.730,
                    },
                    {
                        "featureName": "Sepal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.073,
                        "impactNormalized": 0.241,
                    },
                ],
                "class": "versicolor",
            },
            {
                "featureImpacts": [
                    {
                        "featureName": "Petal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.273,
                        "impactNormalized": 1.0,
                    },
                    {
                        "featureName": "Petal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.174,
                        "impactNormalized": 0.640,
                    },
                    {
                        "featureName": "Sepal Width",
                        "redundantWith": None,
                        "impactUnnormalized": -0.001,
                        "impactNormalized": -0.006,
                    },
                    {
                        "featureName": "Sepal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.018,
                        "impactNormalized": 0.067,
                    },
                ],
                "class": "virginica",
            },
        ],
    }


@pytest.fixture
def feature_impact_response(feature_impact_server_data, feature_impact_url):
    body = json.dumps(feature_impact_server_data)
    responses.add(
        responses.GET, feature_impact_url, status=200, content_type="application/json", body=body
    )


@pytest.fixture
def multiclass_feature_impact_response(
    multiclass_feature_impact_server_data, multiclass_feature_impact_url
):
    body = json.dumps(multiclass_feature_impact_server_data)
    responses.add(
        responses.GET,
        multiclass_feature_impact_url,
        status=200,
        content_type="application/json",
        body=body,
    )


@pytest.fixture
def feature_impact_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType="featureImpact")


@pytest.fixture
def feature_impact_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType="featureImpact")


@pytest.fixture
def approximate_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_RULESETS)


@pytest.fixture
def approximate_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_RULESETS)


@pytest.fixture
def prime_validation_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_VALIDATION)


@pytest.fixture
def prime_validation_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_VALIDATION)


@pytest.fixture
def feature_impact_completed_response(
    feature_impact_job_finished_server_data, job_url, feature_impact_url
):
    """
    Loads a response that the given job is a featureImpact job, and is in completed
    """
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(feature_impact_job_finished_server_data),
        status=303,
        adding_headers={"Location": feature_impact_url},
        content_type="application/json",
    )


@pytest.fixture
def feature_impact_previously_ran_response(project_url, model_id, job_id):
    return _feature_impact_previously_ran_response(project_url, model_id, job_id)


@pytest.fixture
def feature_impact_previously_ran_response__no_job_id(project_url, model_id):
    return _feature_impact_previously_ran_response(
        project_url,
        model_id,
        message="Feature Impact has been computed for this model.",
        job_id=None,
    )


def _feature_impact_previously_ran_response(project_url, model_id, job_id, message=None):
    """
    Loads a response that the given model has already ran its feature impact
     """
    body = {
        "message": message or "Feature Impact is in progress for this model.",
        "errorName": "JobAlreadyAdded",
        "jobId": job_id,
    }
    responses.add(
        responses.POST,
        "{}models/{}/featureImpact/".format(project_url, model_id),
        body=json.dumps(body),
        status=422,
        content_type="application/json",
    )


@pytest.fixture
def feature_impact_running_response(feature_impact_job_running_server_data, job_url):
    """
    Loads a response that the given job is a featureImpact job, and is running
    """
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(feature_impact_job_running_server_data),
        status=200,
        content_type="application/json",
    )


@pytest.fixture
def approximate_job_creation_response(project_url, model_id, job_url):
    responses.add(
        responses.POST,
        "{}models/{}/primeRulesets/".format(project_url, model_id),
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def approximate_completed_response(
    approximate_job_finished_server_data,
    job_url,
    project_url,
    model_id,
    ruleset_without_model_server_data,
    ruleset_with_model_server_data,
):
    rulesets_url = "{}models/{}/primeRulesets/".format(project_url, model_id)
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(approximate_job_finished_server_data),
        status=303,
        adding_headers={"Location": rulesets_url},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        rulesets_url,
        body=json.dumps([ruleset_with_model_server_data, ruleset_without_model_server_data]),
        content_type="application/json",
    )


@pytest.fixture
def prime_validation_job_creation_response(project_url, job_url):
    responses.add(
        responses.POST,
        "{}primeFiles/".format(project_url),
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def prime_validation_job_completed_response(
    prime_validation_job_finished_server_data, job_url, project_url, prime_file_server_data
):
    file_url = "{}primeFiles/{}/".format(project_url, prime_file_server_data["id"])
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(prime_validation_job_finished_server_data),
        status=303,
        adding_headers={"Location": file_url},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        file_url,
        body=json.dumps(prime_file_server_data),
        content_type="application/json",
    )


@pytest.fixture
def mocked_execution_environment_versions():
    return {
        "count": 5,
        "totalCount": 5,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "",
                "buildStatus": "success",
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18a2a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "",
                "buildStatus": "processing",
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18a3a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "some description",
                "buildStatus": "failed",
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18a3a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "some description",
                "buildStatus": "failed",
                "created": "2019-09-28T15:19:26.587583Z",
                "dockerContextSize": 1024,
                "dockerImageSize": 2048,
            },
            {
                "id": "5cf4d3f5f930e26daac18a3a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "some description",
                "buildStatus": "failed",
                "created": "2019-09-28T15:19:26.587583Z",
                "dockerContextSize": None,
                "dockerImageSize": None,
            },
        ],
    }


@pytest.fixture
def mocked_execution_environment_version(mocked_execution_environment_versions):
    return mocked_execution_environment_versions["data"][0]


@pytest.fixture
def mocked_execution_environments(mocked_execution_environment_version):
    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "name": "env1",
                "description": "some description",
                "programmingLanguage": "python",
                "isPublic": True,
                "created": "2019-09-28T15:19:26.587583Z",
                "latestVersion": mocked_execution_environment_version,
            },
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "name": "env2",
                "description": "",
                "programmingLanguage": "other",
                "isPublic": False,
                "created": "2019-09-28T15:19:26.587583Z",
                "latestVersion": None,
            },
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "name": "env2",
                "description": "",
                "programmingLanguage": "other",
                "isPublic": False,
                "created": "2019-09-28T15:19:26.587583Z",
                "latestVersion": None,
                "requiredMetadataKeys": [
                    {"fieldName": "FIELD_{}".format(i), "displayName": "field {}".format(i)}
                    for i in range(3)
                ],
            },
        ],
    }


@pytest.fixture
def mocked_execution_environment(mocked_execution_environments):
    return mocked_execution_environments["data"][0]


@pytest.fixture
def mocked_execution_environment_with_required_metadata_keys(mocked_execution_environments):
    return next(
        env for env in mocked_execution_environments["data"] if env.get("requiredMetadataKeys")
    )


@pytest.fixture
def mocked_custom_model_version_with_required_metadata(mocked_versions):
    return next(version for version in mocked_versions["data"] if version.get("requiredMetadata"))


@pytest.fixture
def shap_impact_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/shapImpact/".format(project_id, model_id)


@pytest.fixture
def shap_impact_job_creation_response(shap_impact_url, job_url):
    responses.add(
        responses.POST, shap_impact_url, body="", status=202, adding_headers={"Location": job_url}
    )


@pytest.fixture
def shap_impact_response(shap_impact_server_data, shap_impact_url):
    body = json.dumps(shap_impact_server_data)
    responses.add(
        responses.GET, shap_impact_url, status=200, content_type="application/json", body=body
    )


@pytest.fixture
def shap_impact_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType="shapImpact")


@pytest.fixture
def shap_impact_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType="shapImpact")


@pytest.fixture
def shap_impact_running_response(shap_impact_job_running_server_data, job_url):
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(shap_impact_job_running_server_data),
        status=200,
        content_type="application/json",
    )


@pytest.fixture
def shap_impact_completed_response(shap_impact_job_finished_server_data, job_url, shap_impact_url):
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(shap_impact_job_finished_server_data),
        status=303,
        adding_headers={"Location": shap_impact_url},
        content_type="application/json",
    )


@pytest.fixture
def shap_impact_previously_ran_response(project_url, model_id, job_id, job_url):
    body = {
        "message": "SHAP impact already exists.",
        "errorName": "JobAlreadyAdded",
        "jobId": job_id,
    }
    responses.add(
        responses.POST,
        "{}models/{}/shapImpact/".format(project_url, model_id),
        body=json.dumps(body),
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )


# CustomTask and CustomTaskVersion


@pytest.fixture
def custom_task_id():
    return "60dcbf0e8cd36258f9f5fff7"


@pytest.fixture
def custom_task_version_id():
    return "60dcbf2b8cd36258f9f5fff8"


@pytest.fixture
def custom_task_version(custom_task_version_id, custom_task_id):
    return {
        "id": custom_task_version_id,
        "customTaskId": custom_task_id,
        "created": "abc",
        "description": "this task is the coolest",
        "items": [],
        "isFrozen": True,
        "versionMajor": 1,
        "versionMinor": 2,
        "label": "1.2",
        "trainingHistory": [],
    }


@pytest.fixture
def workspace_items():
    return [
        {
            "fileSource": "local",
            "workspaceId": "60abd24fc3d97e6f5f732575",
            "filePath": "custom.py",
            "storagePath": "workspace/60abd24fc3/custom.py",
            "repositoryLocation": None,
            "repositoryName": None,
            "repositoryFilePath": None,
            "ref": None,
            "commitSha": None,
            "id": "60abd2abb66ce93dc37324d2",
            "created": "2021-05-24T16:22:03.423629Z",
            "fileName": "custom.py",
        },
        {
            "fileSource": "local",
            "workspaceId": "60abd24fc3d97e6f5f732575",
            "filePath": "createPipeline.py",
            "storagePath": "workspace/60abd24fc3/createPipeline.py",
            "repositoryLocation": "some location",
            "repositoryName": "erics-awesome-repo",
            "repositoryFilePath": "x/y.py",
            "ref": "no idea",
            "commitSha": "123abc",
            "id": "60abd2abb66ce93dc37324d3",
            "created": "2021-05-24T16:22:03.423995Z",
            "fileName": "createPipeline.py",
        },
    ]


@pytest.fixture
def training_history():
    return [
        {"projectId": "def", "modelId": "abc"},
        {"projectId": "ghi", "modelId": "jkl"},
    ]


@pytest.fixture
def custom_task_version_arguments():
    return [
        {
            "key": "user_param___param_float",
            "argument": {
                "name": "param_float",
                "type": "float",
                "default": 0.5,
                "values": [0.0, 1.0],
            },
        },
        {
            "key": "user_param___param_floatgrid",
            "argument": {
                "name": "param_floatgrid",
                "type": "floatgrid",
                "default": 0.5,
                "values": [0.0, 1.0],
            },
        },
        {
            "key": "user_param___param_int",
            "argument": {"name": "param_int", "type": "int", "default": 1, "values": [0, 1]},
        },
        {
            "key": "user_param___param_intgrid",
            "argument": {
                "name": "param_intgrid",
                "type": "intgrid",
                "default": 1,
                "values": [0, 1],
            },
        },
        {
            "key": "user_param___param_multi",
            "argument": {
                "name": "param_multi",
                "type": "multi",
                "default": 1,
                "values": {
                    "int": [0, 1],
                    "float": [0.0, 1.0],
                    "select": {"values": ["a", "b", "c"]},
                },
            },
        },
        {
            "key": "user_param___param_sel",
            "argument": {
                "name": "param_sel",
                "type": "select",
                "default": "b",
                "values": ["a", "b", "c"],
            },
        },
        {
            "key": "user_param___param_selgrid",
            "argument": {
                "name": "param_selgrid",
                "type": "selectgrid",
                "default": "b",
                "values": ["a", "b", "c"],
            },
        },
        {
            "key": "user_param___param_string",
            "argument": {
                "name": "param_string",
                "type": "unicode",
                "default": "default string",
                "values": [],
            },
        },
    ]


@pytest.fixture
def custom_task_version_with_optional_values(
    custom_task_version, workspace_items, training_history, custom_task_version_arguments,
):
    new_task_version = deepcopy(custom_task_version)
    new_task_version.update(
        {
            "baseEnvironmentVersionId": "mno",
            "baseEnvironmentId": "pqr",
            "dependencies": [
                {
                    "packageName": "pandas",
                    "constraints": [{"constraintType": ">=", "version": "1.0"}],
                    "line": "pandas >= 1.0",
                    "lineNumber": 1,
                }
            ],
            "items": workspace_items,
            "trainingHistory": training_history,
            "requiredMetadata": {"hello": "world"},
            "requiredMetadataValues": [{"fieldName": "hello", "value": "world"}],
            "arguments": custom_task_version_arguments,
        }
    )
    return new_task_version


@pytest.fixture
def custom_task_version_response_json(custom_task_version_with_optional_values):
    return deepcopy(custom_task_version_with_optional_values)


@pytest.fixture
def custom_task_version_list_response(custom_task_version_response_json):
    second_version = {
        k: change_response_value(v) for k, v in custom_task_version_response_json.items()
    }

    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "data": [custom_task_version_response_json, second_version],
    }


@pytest.fixture
def custom_task(custom_task_id):
    return {
        "created": "10-10-2020",
        "createdBy": "a froody dude",
        "deploymentsCount": 1,
        "description": "there's some stuff",
        "id": custom_task_id,
        "language": "python",
        "name": "the coolest model",
        "targetType": CUSTOM_MODEL_TARGET_TYPE.BINARY,
        "updated": "12-12-3030",
        "latestVersion": None,
        "customModelType": "training",
    }


@pytest.fixture
def custom_task_with_optional_values(custom_task, custom_task_version_with_optional_values):
    full_task = deepcopy(custom_task)
    full_task["latestVersion"] = deepcopy(custom_task_version_with_optional_values)
    full_task["calibratePredictions"] = False
    return full_task


@pytest.fixture
def custom_task_response_json(custom_task_with_optional_values):
    return deepcopy(custom_task_with_optional_values)


@pytest.fixture
def custom_task_list_response(custom_task_response_json):
    second = {k: change_response_value(v) for k, v in custom_task_response_json.items()}
    second["latestVersion"] = None

    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "data": [custom_task_response_json, second],
    }


@pytest.fixture
def custom_task_version_dependency_build_server_data(custom_task_id, custom_task_version_id):
    return {
        "customTaskId": custom_task_id,
        "customTaskVersionId": custom_task_version_id,
        "buildStart": "2018-11-19T14:18:01.468272Z",
        "buildEnd": "2018-11-19T15:18:01.468272Z",
        "buildStatus": "success",
        "buildLogLocation": "https://127.0.0.1/logs/",
    }



# -*- coding: utf-8 -*-
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from copy import deepcopy
import json

import numpy as np
import pandas as pd
import pytest
import responses
import trafaret

from datarobot import errors
from datarobot.models.shap_matrix import ShapMatrix
from tests.utils import add_response, request_body_to_json


@pytest.fixture
def shap_matrix_id():
    return "5ea6e6ab4cfad16f3499940e"


@pytest.fixture
def mocked_shap_matrix_values():
    row_indexes = [u"row_{}".format(i) for i in range(10)]
    column_names = [u"カテゴリー_{}".format(i) for i in range(5)]
    rs = np.random.RandomState(1234)
    shap_scores = rs.uniform(-1, 1, (len(row_indexes), len(column_names)))
    df = pd.DataFrame(data=shap_scores, columns=column_names, index=row_indexes)
    return df.to_csv(encoding="utf-8")


@pytest.fixture
def mocked_shap_matrices(shap_matrix_id, project_id, model_id, dataset_id):
    shap_matrices_json = [
        {
            "id": shap_matrix_id,
            "projectId": project_id,
            "modelId": model_id,
            "datasetId": dataset_id,
        },
    ]
    return {
        "count": len(shap_matrices_json),
        "next": None,
        "previous": None,
        "data": shap_matrices_json,
    }


@pytest.fixture
def mocked_shap_matrix(mocked_shap_matrices):
    return mocked_shap_matrices["data"][0]


@pytest.fixture
def make_shap_matrices_url(unittest_endpoint, project_id):
    def _make_shap_matrices_url(shap_matrix_id=None):
        base_url = unittest_endpoint + "/projects/{}/shapMatrices/".format(project_id)
        if shap_matrix_id is not None:
            return "{}{}/".format(base_url, shap_matrix_id)
        return base_url

    return _make_shap_matrices_url


def assert_shap_matrix(shap_matrix, shap_matrix_json):
    assert shap_matrix.id == shap_matrix_json["id"]
    assert shap_matrix.project_id == shap_matrix_json["projectId"]
    assert shap_matrix.model_id == shap_matrix_json["modelId"]
    assert shap_matrix.dataset_id == shap_matrix_json["datasetId"]


def test_from_server_data(mocked_shap_matrix):
    shap_matrix = ShapMatrix.from_server_data(mocked_shap_matrix)
    assert_shap_matrix(shap_matrix, mocked_shap_matrix)


@responses.activate
def test_get__no_request_made(project_id, shap_matrix_id):
    ShapMatrix.get(project_id, shap_matrix_id)
    assert len(responses.calls) == 0


@responses.activate
def test_list(project_id, mocked_shap_matrices, make_shap_matrices_url):
    url = make_shap_matrices_url()
    add_response(url, mocked_shap_matrices)

    shap_matrices = ShapMatrix.list(project_id)
    assert shap_matrices is not None
    assert len(mocked_shap_matrices["data"]) == len(shap_matrices)
    for shap_matrix, shap_matrix_json in zip(shap_matrices, mocked_shap_matrices["data"]):
        assert_shap_matrix(shap_matrix, shap_matrix_json)
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_multiple_pages(project_id, mocked_shap_matrices, make_shap_matrices_url):
    url1 = make_shap_matrices_url()
    url2 = make_shap_matrices_url() + "2"

    mocked_shap_matrices_2nd = deepcopy(mocked_shap_matrices)
    mocked_shap_matrices["next"] = url2

    add_response(url1, mocked_shap_matrices)
    add_response(url2, mocked_shap_matrices_2nd)

    shap_matrices = ShapMatrix.list(project_id)
    assert len(shap_matrices) == len(mocked_shap_matrices["data"]) + len(
        mocked_shap_matrices_2nd["data"]
    )
    for shap_matrix, shap_matrix_json in zip(
        shap_matrices, mocked_shap_matrices["data"] + mocked_shap_matrices_2nd["data"]
    ):
        assert_shap_matrix(shap_matrix, shap_matrix_json)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith(url1)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(url2)


@responses.activate
def test_list__no_id_in_api_response_causes_error(
    project_id, mocked_shap_matrices, make_shap_matrices_url
):
    mocked_shap_matrices = deepcopy(mocked_shap_matrices)
    shap_matrix_no_id = {
        "id": None,
        "projectId": project_id,
        "modelId": "model-id",
        "datasetId": "dataset-id",
    }
    mocked_shap_matrices["data"].append(shap_matrix_no_id)

    url = make_shap_matrices_url()
    add_response(url, mocked_shap_matrices)

    with pytest.raises(trafaret.DataError):
        ShapMatrix.list(project_id)


@responses.activate
def test_get_as_dataframe(
    project_id, shap_matrix_id, mocked_shap_matrix_values, make_shap_matrices_url
):
    url = make_shap_matrices_url(shap_matrix_id)
    add_response(url, mocked_shap_matrix_values, content_type="text/csv")

    shap_matrix = ShapMatrix.get(project_id, shap_matrix_id)
    shap_matrix_values = shap_matrix.get_as_dataframe()
    assert isinstance(shap_matrix_values, pd.DataFrame)
    assert shap_matrix_values.shape == (10, 5)


@responses.activate
def test_get_as_dataframe__not_found(
    project_id, shap_matrix_id, mocked_shap_matrix_values, make_shap_matrices_url
):
    url = make_shap_matrices_url(shap_matrix_id)
    add_response(url, "", status=404)

    shap_matrix = ShapMatrix.get(project_id, shap_matrix_id)
    with pytest.raises(errors.ClientError):
        shap_matrix.get_as_dataframe()


@responses.activate
def test_get_as_dataframe__unknown_error(
    project_id, shap_matrix_id, mocked_shap_matrix_values, make_shap_matrices_url
):
    url = make_shap_matrices_url(shap_matrix_id)
    add_response(url, "", status=500)

    shap_matrix = ShapMatrix.get(project_id, shap_matrix_id)
    with pytest.raises(errors.ServerError):
        shap_matrix.get_as_dataframe()


@responses.activate
def test_create(one_model, project_url, base_job_completed_server_data):
    shap_matrices_url = project_url + "shapMatrices/"
    job_data = dict(base_job_completed_server_data, jobType="shapMatrix")
    job_id = job_data["id"]
    job_url = project_url + "jobs/{}/".format(job_id)
    dataset_id = "12344321beefcafe43211234"
    shap_matrix_id = "deadbeef12344321feebdaed"
    finished_shap_matrix_url = shap_matrices_url + "{}/".format(shap_matrix_id)

    submit_shap_job_request = {"modelId": one_model.id, "datasetId": dataset_id}
    responses.add(
        responses.POST,
        shap_matrices_url,
        body=json.dumps(submit_shap_job_request),
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )

    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(job_data),
        status=303,
        content_type="application/json",
        adding_headers={"Location": finished_shap_matrix_url},
    )

    shap_matrix_job = ShapMatrix.create(
        project_id=one_model.project_id, model_id=one_model.id, dataset_id=dataset_id
    )
    assert shap_matrix_job.id == int(job_id)

    request_json = request_body_to_json(responses.calls[0].request)
    assert request_json.get("modelId") == submit_shap_job_request["modelId"]
    assert request_json.get("datasetId") == submit_shap_job_request["datasetId"]


@responses.activate
def test_create__client_error(project_url, one_model, dataset_id):
    error_message = {"message": "some error message"}
    responses.add(
        responses.POST,
        project_url + "shapMatrices/",
        body=json.dumps(error_message),
        status=422,
        content_type="application/json",
    )
    with pytest.raises(errors.ClientError):
        ShapMatrix.create(
            project_id=one_model.project_id, model_id=one_model.id, dataset_id=dataset_id
        )


@responses.activate
def test_create__server_error(project_url, one_model, dataset_id):
    shap_matrices_url = project_url + "shapMatrices/"
    submit_shap_job_request = {"modelId": one_model.id, "datasetId": dataset_id}
    responses.add(
        responses.POST,
        shap_matrices_url,
        body=json.dumps(submit_shap_job_request),
        status=500,
        content_type="application/json",
    )
    with pytest.raises(errors.ServerError):
        ShapMatrix.create(
            project_id=one_model.project_id, model_id=one_model.id, dataset_id=dataset_id
        )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from tempfile import NamedTemporaryFile

import pytest
import responses

from datarobot import Deployment


@pytest.fixture
def scoring_code_data(deployment_data):
    return b"scoring_code_jar_file_data"


@pytest.fixture
def scoring_code_retrieve_response(unittest_endpoint, deployment_data, scoring_code_data):
    url = "{}/deployments/{}/scoringCode/".format(unittest_endpoint, deployment_data["id"])
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/java-archive",
        body=scoring_code_data,
    )


@pytest.fixture
def scoring_code_build_response(unittest_endpoint, deployment_data):
    status_url = "{}/status_url".format(unittest_endpoint)
    scoring_code_build_url = "{}/deployments/{}/scoringCodeBuilds/".format(
        unittest_endpoint, deployment_data["id"]
    )
    scoring_code_retrieve_url = "{}/deployments/{}/scoringCode/".format(
        unittest_endpoint, deployment_data["id"]
    )

    responses.add(
        responses.GET, status_url, headers={"Location": scoring_code_retrieve_url}, status=303
    )
    responses.add(
        responses.POST, scoring_code_build_url, headers={"Location": status_url}, status=202
    )


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "scoring_code_retrieve_response")
def test_download_scoring_code_without_build(deployment_data, scoring_code_data):
    """Test download_scoring_code that does not require building."""

    with NamedTemporaryFile() as file:
        deployment = Deployment.get(deployment_data["id"])
        deployment.download_scoring_code(file.name, include_agent=False)
        assert file.read() == scoring_code_data


@responses.activate
@pytest.mark.usefixtures(
    "deployment_get_response", "scoring_code_build_response", "scoring_code_retrieve_response"
)
def test_download_scoring_code_with_build(deployment_data, scoring_code_data):
    """Test download_scoring_code that requires building first."""

    with NamedTemporaryFile() as file:
        deployment = Deployment.get(deployment_data["id"])
        deployment.download_scoring_code(file.name, include_agent=True)
        assert file.read() == scoring_code_data



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import responses

from datarobot import RelationshipsConfiguration
from datarobot.helpers.feature_discovery import FeatureDiscoverySetting


@responses.activate
def test_relationships_config_creation(
    relationships_configuration,
    create_dataset_definitions,
    create_relationships,
    create_feature_discovery_settings,
):
    responses.add(
        responses.POST,
        "https://host_name.com/relationshipsConfigurations/",
        json=relationships_configuration,
    )

    settings = [
        FeatureDiscoverySetting(s["name"], s["value"]) for s in create_feature_discovery_settings
    ]
    relationships_configuration = RelationshipsConfiguration.create(
        dataset_definitions=create_dataset_definitions,
        relationships=create_relationships,
        feature_discovery_settings=settings,
    )
    assert responses.calls[0].request.method == "POST"
    assert relationships_configuration.id == "5a530498d5c1f302d6d176c8"
    assert len(relationships_configuration.dataset_definitions) == 2
    assert len(relationships_configuration.relationships) == 2
    assert len(relationships_configuration.feature_discovery_settings) == 2
    assert relationships_configuration.feature_discovery_mode == "manual"


@responses.activate
def test_relationships_config_retrieve(
    relationships_configuration, create_dataset_definitions, create_relationships
):
    id = relationships_configuration["id"]
    responses.add(
        responses.GET,
        "https://host_name.com/relationshipsConfigurations/{}/".format(id),
        json=relationships_configuration,
    )

    result = RelationshipsConfiguration(id=id).get()

    assert responses.calls[0].request.method == "GET"
    assert result.id == id
    assert len(result.dataset_definitions) == 2
    assert len(result.relationships) == 2
    for actual, expected in zip(result.dataset_definitions, create_dataset_definitions):
        assert actual["identifier"] == expected["identifier"]
        assert actual["catalog_version_id"] == expected["catalogVersionId"]
        assert actual["catalog_id"] == expected["catalogId"]
        assert actual["snapshot_policy"] == expected["snapshotPolicy"]

    actual = result.relationships[1]
    expected = create_relationships[1]
    assert actual["dataset1_identifier"] == expected["dataset1Identifier"]
    assert actual["dataset2_identifier"] == expected["dataset2Identifier"]
    assert actual["dataset1_keys"] == expected["dataset1Keys"]
    assert actual["dataset2_keys"] == expected["dataset2Keys"]
    assert actual["feature_derivation_window_end"] == expected["featureDerivationWindowEnd"]
    assert actual["feature_derivation_window_start"] == expected["featureDerivationWindowStart"]
    assert (
        actual["feature_derivation_window_time_unit"] == expected["featureDerivationWindowTimeUnit"]
    )
    assert actual["prediction_point_rounding"] == expected["predictionPointRounding"]
    assert (
        actual["prediction_point_rounding_time_unit"] == expected["predictionPointRoundingTimeUnit"]
    )


@responses.activate
def test_relationships_config_replace(
    relationships_configuration,
    create_dataset_definitions,
    create_relationships,
    create_feature_discovery_settings,
):
    responses.add(
        responses.PUT,
        "https://host_name.com/relationshipsConfigurations/R-ID/",
        json=relationships_configuration,
    )

    relationships_configuration = RelationshipsConfiguration("R-ID").replace(
        dataset_definitions=create_dataset_definitions,
        relationships=create_relationships,
        feature_discovery_settings=create_feature_discovery_settings,
    )
    assert responses.calls[0].request.method == "PUT"
    assert relationships_configuration.id == "5a530498d5c1f302d6d176c8"
    assert len(relationships_configuration.dataset_definitions) == 2
    assert len(relationships_configuration.relationships) == 2
    assert len(relationships_configuration.feature_discovery_settings) == 2
    assert relationships_configuration.feature_discovery_mode == "manual"


@responses.activate
def test_relationships_config_delete(relationships_configuration):
    id = relationships_configuration["id"]
    url = "https://host_name.com/relationshipsConfigurations/{}/".format(id)
    responses.add(responses.DELETE, url)

    relationships_config = RelationshipsConfiguration(id=id)
    relationships_config.delete()

    assert responses.calls[0].request.method == responses.DELETE
    assert responses.calls[0].request.url == url


@responses.activate
def test_jdbc_relationships_config_creation(
    jdbc_relationships_configuration, create_jdbc_dataset_definitions, create_relationships
):
    responses.add(
        responses.POST,
        "https://host_name.com/relationshipsConfigurations/",
        json=jdbc_relationships_configuration,
    )

    relationships_configuration = RelationshipsConfiguration.create(
        dataset_definitions=create_jdbc_dataset_definitions, relationships=create_relationships,
    )
    assert responses.calls[0].request.method == "POST"
    assert relationships_configuration.id == "5a530498d5c1f302d6d176c8"
    assert len(relationships_configuration.dataset_definitions) == 1
    assert len(relationships_configuration.relationships) == 2
    definition = relationships_configuration.dataset_definitions[0]
    assert definition["data_source"]["data_store_name"] == "mssql-remote-dev"
    assert definition["data_source"]["data_store_id"] == "60c6e3f833de57c65440f571"
    assert definition["data_source"]["data_source_id"] == "60c6e3f833de57c65440f581"
    assert definition["data_source"]["dbtable"] == "safer_credit"
    assert definition["data_source"]["schema"] == "dbo"
    assert definition["data_source"]["catalog"] == "safer"


@responses.activate
def test_relationships_config_with_multiple_fdws_creation(
    relationships_configuration_multiple_fdws,
    create_dataset_definitions,
    create_relationships_multiple_fdws,
    create_feature_discovery_settings,
):
    responses.add(
        responses.POST,
        "https://host_name.com/relationshipsConfigurations/",
        json=relationships_configuration_multiple_fdws,
    )

    settings = [
        FeatureDiscoverySetting(s["name"], s["value"]) for s in create_feature_discovery_settings
    ]
    relationships_configuration = RelationshipsConfiguration.create(
        dataset_definitions=create_dataset_definitions,
        relationships=create_relationships_multiple_fdws,
        feature_discovery_settings=settings,
    )
    assert responses.calls[0].request.method == "POST"
    assert relationships_configuration.id == "5a530498d5c1f302d6d176c8"
    assert len(relationships_configuration.dataset_definitions) == 2
    assert len(relationships_configuration.relationships) == 2
    assert relationships_configuration.relationships[1]["feature_derivation_windows"] == [
        {"end": -1, "start": -10, "unit": u"SECOND"},
        {"end": -10, "start": -20, "unit": u"SECOND"},
    ]
    assert len(relationships_configuration.feature_discovery_settings) == 2
    assert relationships_configuration.feature_discovery_mode == "manual"


@responses.activate
def test_relationships_config_multiple_fdws_retrieve(
    relationships_configuration_multiple_fdws,
    create_dataset_definitions,
    create_relationships_multiple_fdws,
):
    id = relationships_configuration_multiple_fdws["id"]
    responses.add(
        responses.GET,
        "https://host_name.com/relationshipsConfigurations/{}/".format(id),
        json=relationships_configuration_multiple_fdws,
    )

    result = RelationshipsConfiguration(id=id).get()

    assert responses.calls[0].request.method == "GET"
    assert result.id == id
    assert len(result.dataset_definitions) == 2
    assert len(result.relationships) == 2
    for actual, expected in zip(result.dataset_definitions, create_dataset_definitions):
        assert actual["identifier"] == expected["identifier"]
        assert actual["catalog_version_id"] == expected["catalogVersionId"]
        assert actual["catalog_id"] == expected["catalogId"]
        assert actual["snapshot_policy"] == expected["snapshotPolicy"]

    actual = result.relationships[1]
    expected = create_relationships_multiple_fdws[1]
    assert actual["dataset1_identifier"] == expected["dataset1Identifier"]
    assert actual["dataset2_identifier"] == expected["dataset2Identifier"]
    assert actual["dataset1_keys"] == expected["dataset1Keys"]
    assert actual["dataset2_keys"] == expected["dataset2Keys"]
    assert actual["feature_derivation_windows"] == expected["featureDerivationWindows"]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import copy
from copy import deepcopy
import json
import os
from shutil import rmtree

import mock
import pytest
from requests_toolbelt import MultipartEncoder
import responses
from trafaret import DataError

from datarobot import errors
from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.models.custom_model_version import RequiredMetadataValue
from datarobot.models.custom_task_version import CustomTaskFileItem, CustomTaskVersion
from datarobot.utils import camelize, underscorize
from tests.model.utils import (
    assert_custom_task_version,
    assert_custom_task_version_dependency_build,
    assert_items,
    assert_single_api_call,
    change_response_value,
    fields_to_dict,
)


@pytest.fixture
def version_url(unittest_endpoint, custom_task_id, custom_task_version_id):
    return "{}/customTasks/{}/versions/{}/".format(
        unittest_endpoint, custom_task_id, custom_task_version_id
    )


@pytest.fixture
def versions_base_url(unittest_endpoint, custom_task_id):
    return "{}/customTasks/{}/versions/".format(unittest_endpoint, custom_task_id)


@pytest.fixture
def nullable_keys():
    return {
        "description",
        "baseEnvironmentId",
        "baseEnvironmentVersionId",
    }


def add_mock_get_response(url, json_dict):
    add_mock_response(responses.GET, url, json.dumps(json_dict))


def add_mock_response(method, url, body):
    responses.add(method, url, status=200, content_type="application/json", body=body)


@pytest.fixture()
def version_get_response(version_url, custom_task_version_response_json):
    add_mock_get_response(version_url, custom_task_version_response_json)


@pytest.fixture
def dependency_img_build_route(unittest_endpoint, custom_task_id, custom_task_version_id):
    return "{}/customTasks/{}/versions/{}/dependencyBuild/".format(
        unittest_endpoint, custom_task_id, custom_task_version_id,
    )


@pytest.fixture
def dependency_img_build_log_route(unittest_endpoint, custom_task_id, custom_task_version_id):
    return "{}/customTasks/{}/versions/{}/dependencyBuildLog/".format(
        unittest_endpoint, custom_task_id, custom_task_version_id,
    )


FILE_ITEM_RESPONSE_KEYS = [camelize(el.name) for el in CustomTaskFileItem.schema.keys]


class TestCustomTaskFileItem:
    def test_from_server_data_custom_model_file_item(self, workspace_items):
        assert_items(
            [CustomTaskFileItem.from_server_data(el) for el in workspace_items], workspace_items
        )

    @pytest.mark.parametrize("key", FILE_ITEM_RESPONSE_KEYS)
    def test_from_server_data_no_optional_values(self, workspace_items, key):
        """Necessary because the class it inherits from allows created to be nullable."""
        item = workspace_items[0]
        item[key] = None
        with pytest.raises(DataError) as exec_info:
            CustomTaskFileItem.from_server_data(item)
        error = exec_info.value
        assert len(error.error) == 1
        assert error.error[underscorize(key)].error == "is required"


class TestCustomTaskVersionGetActionsAndUpdate:
    def test_from_server_data_minimal_response(self, custom_task_version):
        version = CustomTaskVersion.from_server_data(custom_task_version)
        assert_custom_task_version(version, custom_task_version)

    def test_from_server_data_full_response(self, custom_task_version_with_optional_values):
        version = CustomTaskVersion.from_server_data(custom_task_version_with_optional_values)
        assert_custom_task_version(version, custom_task_version_with_optional_values)

    def test_from_server_data_nullable_values(
        self, custom_task_version_response_json, nullable_keys
    ):
        for key in nullable_keys:
            custom_task_version_response_json[key] = None
        version = CustomTaskVersion.from_server_data(custom_task_version_response_json)
        assert_custom_task_version(version, custom_task_version_response_json)

    @responses.activate
    @pytest.mark.usefixtures("version_get_response")
    def test_get_custom_task_version(self, custom_task_version_response_json):
        task_id = custom_task_version_response_json["customTaskId"]
        version_id = custom_task_version_response_json["id"]
        version = CustomTaskVersion.get(task_id, version_id)
        assert_custom_task_version(version, custom_task_version_response_json)

        assert_single_api_call(
            responses.calls,
            responses.GET,
            "/customTasks/{}/versions/{}/".format(task_id, version_id),
        )

    @responses.activate
    def test_refresh_custom_task_version(self, custom_task_version_response_json, version_url):
        task_id = custom_task_version_response_json["customTaskId"]
        version_id = custom_task_version_response_json["id"]

        updated_response = {
            key: change_response_value(value)
            for key, value in custom_task_version_response_json.items()
        }

        add_mock_get_response(version_url, custom_task_version_response_json)
        add_mock_get_response(version_url, updated_response)

        version = CustomTaskVersion.get(task_id, version_id)

        assert len(responses.calls) == 1
        assert_custom_task_version(version, custom_task_version_response_json)

        version.refresh()

        assert len(responses.calls) == 2
        assert_custom_task_version(version, updated_response)

    @responses.activate
    def test_list_response_single_list(
        self, custom_task_version_list_response, custom_task_id, versions_base_url
    ):
        add_mock_get_response(versions_base_url, custom_task_version_list_response)
        result = CustomTaskVersion.list(custom_task_id)
        assert len(result) == len(custom_task_version_list_response["data"])
        for actual, expected in zip(result, custom_task_version_list_response["data"]):
            assert_custom_task_version(actual, expected)

        assert_single_api_call(
            responses.calls, responses.GET, "/customTasks/{}/versions/".format(custom_task_id)
        )

    @responses.activate
    def test_list_response_multi_list(
        self, custom_task_version_list_response, custom_task_id, versions_base_url
    ):
        next_page = deepcopy(custom_task_version_list_response)
        next_data = next_page["data"]
        for el in next_data:
            for key in el:
                el[key] = change_response_value(el[key])

        next_url = "https://www.coolguy.com/coolness/"
        custom_task_version_list_response["next"] = next_url
        add_mock_get_response(versions_base_url, custom_task_version_list_response)
        add_mock_get_response(next_url, next_page)
        result = CustomTaskVersion.list(custom_task_id)
        assert len(result) == len(custom_task_version_list_response["data"]) * 2
        for actual, expected in zip(result, custom_task_version_list_response["data"] + next_data):
            assert_custom_task_version(actual, expected)

    @responses.activate
    @pytest.mark.usefixtures("version_get_response")
    def test_download(self, temporary_file, version_url, custom_task_id, custom_task_version_id):
        target_url = version_url + "download/"
        content = b"some stuff"
        add_mock_response(responses.GET, target_url, content)
        version = CustomTaskVersion.get(custom_task_id, custom_task_version_id)
        version.download(temporary_file)
        with open(temporary_file, "rb") as f:
            assert f.read() == content

        calls_after_get = responses.calls[1:]
        expected_url = "customTasks/{}/versions/{}/download/".format(
            custom_task_id, custom_task_version_id
        )
        assert_single_api_call(calls_after_get, responses.GET, expected_url)

    @responses.activate
    @pytest.mark.usefixtures("version_get_response")
    def test_update(
        self, version_url, custom_task_id, custom_task_version_id, custom_task_version_response_json
    ):
        updated_json = deepcopy(custom_task_version_response_json)
        updated_json["requiredMetadata"] = {"TEST_UPDATE": "new_value", "otherThing": "OthErValue"}
        updated_json["description"] = "test_update description which is so awesome"

        add_mock_response(responses.PATCH, version_url, body=json.dumps(updated_json))

        version = CustomTaskVersion.get(custom_task_id, custom_task_version_id)
        version.update("dummy_value", {"DUMMY_VALUE": "value"})

        calls_after_get = responses.calls[1:]
        expected_url_ending = "customTasks/{}/versions/{}/".format(
            version.custom_task_id, version.id
        )
        assert_single_api_call(calls_after_get, responses.PATCH, expected_url_ending)

        call = calls_after_get[0]
        assert json.loads(call.request.body) == {
            "description": "dummy_value",
            "requiredMetadata": {"DUMMY_VALUE": "value"},
        }
        assert_custom_task_version(version, updated_json)

    @responses.activate
    @pytest.mark.usefixtures("version_get_response")
    def test_update_only_description(
        self, version_url, custom_task_id, custom_task_version_id, custom_task_version_response_json
    ):
        add_mock_response(
            responses.PATCH, version_url, body=json.dumps(custom_task_version_response_json)
        )

        version = CustomTaskVersion.get(custom_task_id, custom_task_version_id)
        version.update(description="only description")

        call = responses.calls[1]
        assert json.loads(call.request.body) == {"description": "only description"}

    @responses.activate
    @pytest.mark.usefixtures("version_get_response")
    def test_update_with_deprecated_required_metadata(
        self,
        version_url,
        custom_task_id,
        custom_task_version_id,
        custom_task_version_response_json,
    ):
        add_mock_response(
            responses.PATCH, version_url, body=json.dumps(custom_task_version_response_json)
        )

        version = CustomTaskVersion.get(custom_task_id, custom_task_version_id)
        attrs = {"required_metadata": {"hello": "world"}}

        version.update(**attrs)
        call = responses.calls[1]
        assert json.loads(call.request.body) == {"requiredMetadata": {"hello": "world"}}

    @responses.activate
    @pytest.mark.usefixtures("version_get_response")
    def test_update_with_required_metadata_values(
        self,
        version_url,
        custom_task_id,
        custom_task_version_id,
        custom_task_version_response_json,
    ):
        add_mock_response(
            responses.PATCH, version_url, body=json.dumps(custom_task_version_response_json)
        )

        version = CustomTaskVersion.get(custom_task_id, custom_task_version_id)
        attrs = {
            "required_metadata_values": [RequiredMetadataValue(field_name="hello", value="world")]
        }

        version.update(**attrs)
        call = responses.calls[1]
        assert json.loads(call.request.body) == {
            "requiredMetadataValues": [{"fieldName": "hello", "value": "world"}]
        }

    @responses.activate
    @pytest.mark.usefixtures("version_get_response")
    @pytest.mark.parametrize("required_metadata", [{}, {"a": "b"}], ids=["empty", "non-empty"])
    @pytest.mark.parametrize(
        "required_metadata_values",
        [[], [RequiredMetadataValue(field_name="a", value="b")]],
        ids=["empty", "non-empty"],
    )
    def test_update_with_both_required_metadata_and_required_metadata_value_raises_error(
        self, custom_task_id, custom_task_version_id, required_metadata, required_metadata_values
    ):

        version = CustomTaskVersion.get(custom_task_id, custom_task_version_id)
        attrs = {
            "required_metadata_values": required_metadata_values,
            "required_metadata": required_metadata,
        }

        with pytest.raises(ValueError):
            version.update(**attrs)


class TestCreateMethods:
    @staticmethod
    def _get_http_method(create_method):
        if create_method == CustomTaskVersion.create_clean:
            http_method = responses.POST
        else:
            http_method = responses.PATCH
        return http_method

    @responses.activate
    @pytest.mark.parametrize(
        "create_method",
        [CustomTaskVersion.create_clean, CustomTaskVersion.create_from_previous],
        ids=["create_clean", "create_from_previous"],
    )
    def test_minimal_request(
        self, versions_base_url, custom_task_id, custom_task_version_response_json, create_method
    ):

        http_method = self._get_http_method(create_method)

        add_mock_response(
            http_method, versions_base_url, json.dumps(custom_task_version_response_json)
        )

        base_env_id = "abc123"
        response = create_method(custom_task_id, base_env_id)

        assert_single_api_call(
            responses.calls, http_method, "customTasks/{}/versions/".format(custom_task_id)
        )
        assert_custom_task_version(response, custom_task_version_response_json)

        request_body = responses.calls[0].request.body
        assert isinstance(request_body, MultipartEncoder)
        assert fields_to_dict(request_body) == {
            "baseEnvironmentId": [base_env_id],
            "isMajorUpdate": ["True"],
        }

    @responses.activate
    @pytest.mark.parametrize(
        "create_method",
        [CustomTaskVersion.create_clean, CustomTaskVersion.create_from_previous],
        ids=["create_clean", "create_from_previous"],
    )
    @pytest.mark.parametrize("is_major_update", [True, False])
    def test_optional_values_with_deprecated_required_metadata(
        self,
        versions_base_url,
        custom_task_id,
        custom_task_version_response_json,
        create_method,
        is_major_update,
    ):

        http_method = self._get_http_method(create_method)

        add_mock_response(
            http_method, versions_base_url, json.dumps(custom_task_version_response_json)
        )

        base_env_id = "abc123"
        required_metadata = {"HI_THERE": "stuff"}
        attrs = dict(
            custom_task_id=custom_task_id,
            base_environment_id=base_env_id,
            is_major_update=is_major_update,
            required_metadata=required_metadata,
        )
        api_attrs = {
            "baseEnvironmentId": [base_env_id],
            "isMajorUpdate": [str(is_major_update)],
            "requiredMetadata": [json.dumps(required_metadata)],
        }

        create_method(**attrs)

        request_body = responses.calls[0].request.body
        assert fields_to_dict(request_body) == api_attrs
        assert (
            responses.calls[0]
            .request.headers["Content-Type"]
            .startswith("multipart/form-data; boundary=")
        )

    @responses.activate
    @pytest.mark.parametrize(
        "create_method",
        [CustomTaskVersion.create_clean, CustomTaskVersion.create_from_previous],
        ids=["create_clean", "create_from_previous"],
    )
    @pytest.mark.parametrize("is_major_update", [True, False])
    def test_optional_values_with_required_metadata_values(
        self,
        versions_base_url,
        custom_task_id,
        custom_task_version_response_json,
        create_method,
        is_major_update,
    ):

        http_method = self._get_http_method(create_method)

        add_mock_response(
            http_method, versions_base_url, json.dumps(custom_task_version_response_json)
        )

        base_env_id = "abc123"
        required_metadata = {"HI_THERE": "stuff"}
        attrs = dict(
            custom_task_id=custom_task_id,
            base_environment_id=base_env_id,
            is_major_update=is_major_update,
            required_metadata_values=[
                RequiredMetadataValue(field_name=k, value=v) for k, v in required_metadata.items()
            ],
        )
        api_attrs = {
            "baseEnvironmentId": [base_env_id],
            "isMajorUpdate": [str(is_major_update)],
            "requiredMetadataValues": [
                json.dumps(
                    [
                        {camelize(k): v for k, v in val.to_dict().items()}
                        for val in attrs["required_metadata_values"]
                    ]
                )
            ],
        }

        create_method(**attrs)

        request_body = responses.calls[0].request.body
        assert fields_to_dict(request_body) == api_attrs
        assert (
            responses.calls[0]
            .request.headers["Content-Type"]
            .startswith("multipart/form-data; boundary=")
        )

    @pytest.mark.parametrize(
        "create_method",
        [CustomTaskVersion.create_clean, CustomTaskVersion.create_from_previous],
        ids=["create_clean", "create_from_previous"],
    )
    @pytest.mark.parametrize("required_metadata", [{}, {"a": "b"}], ids=["empty", "non-empty"])
    @pytest.mark.parametrize(
        "required_metadata_values",
        [[], [RequiredMetadataValue(field_name="a", value="b")]],
        ids=["empty", "non-empty"],
    )
    def test_optional_values_with_both_required_metadata_and_required_metadata_value_raises_error(
        self, create_method, required_metadata, required_metadata_values
    ):

        attrs = dict(
            custom_task_id="def456",
            base_environment_id="abc123",
            required_metadata_values=required_metadata_values,
            required_metadata=required_metadata,
        )

        with pytest.raises(ValueError):
            create_method(**attrs)

    @responses.activate
    @pytest.mark.parametrize(
        "create_method",
        [CustomTaskVersion.create_clean, CustomTaskVersion.create_from_previous],
        ids=["create_clean", "create_from_previous"],
    )
    def test_file_path(
        self,
        temporary_dir,
        versions_base_url,
        custom_task_id,
        custom_task_version_response_json,
        create_method,
    ):
        top_file_name = "a-file.txt"
        sub_file_name = "sub-file.txt"
        top_file_path = os.path.join(temporary_dir, top_file_name)
        sub_file_path = os.path.join(temporary_dir, "sub-dir", sub_file_name)

        os.mkdir(os.path.dirname(sub_file_path))

        for file_name in (top_file_path, sub_file_path):
            with open(file_name, "wb") as f:
                f.write(b"some content")

        http_method = self._get_http_method(create_method)
        add_mock_response(
            http_method, versions_base_url, json.dumps(custom_task_version_response_json)
        )

        create_method(custom_task_id, "abc", folder_path=temporary_dir)

        fields = fields_to_dict(responses.calls[0].request.body)

        files = fields["file"]
        paths = fields["filePath"]

        parent_dir_size = len(temporary_dir + "/")
        relative_paths = {
            top_file_name: top_file_path[parent_dir_size:],
            sub_file_name: sub_file_path[parent_dir_size:],
        }
        for (filename, file_obj), file_path in zip(files, paths):
            assert relative_paths[filename] == file_path
            assert file_obj.closed

    @responses.activate
    @pytest.mark.parametrize(
        "create_method",
        [CustomTaskVersion.create_clean, CustomTaskVersion.create_from_previous],
        ids=["create_clean", "create_from_previous"],
    )
    def test_bad_file_path_raises_value_error(
        self,
        temporary_dir,
        versions_base_url,
        custom_task_id,
        custom_task_version_response_json,
        create_method,
    ):
        garbage_file_path = os.path.join(temporary_dir, "garbage_dir")
        rmtree(garbage_file_path, ignore_errors=True)

        assert not os.path.exists(garbage_file_path)

        http_method = self._get_http_method(create_method)
        add_mock_response(
            http_method, versions_base_url, json.dumps(custom_task_version_response_json)
        )

        with pytest.raises(ValueError):
            create_method(custom_task_id, "abc", folder_path=garbage_file_path)

    @responses.activate
    def test_create_from_previous_files_to_delete(
        self, versions_base_url, custom_task_id, custom_task_version_response_json,
    ):
        add_mock_response(
            responses.PATCH, versions_base_url, json.dumps(custom_task_version_response_json)
        )

        CustomTaskVersion.create_from_previous(
            custom_task_id, "abc123", files_to_delete=["abc", "def"]
        )
        fields = fields_to_dict(responses.calls[0].request.body)
        assert set(fields["filesToDelete"]) == {"abc", "def"}
        assert len(fields["filesToDelete"]) == 2


class TestDependencyImageBuild:
    @responses.activate
    def test_start_dependency_build__no_waiting(
        self,
        dependency_img_build_route,
        custom_task_version,
        custom_task_version_dependency_build_server_data,
    ):
        version = CustomTaskVersion.from_server_data(custom_task_version)
        responses.add(
            responses.POST,
            dependency_img_build_route,
            status=202,
            content_type="application/json",
            json=custom_task_version_dependency_build_server_data,
        )

        dependency_build = version.start_dependency_build()
        assert len(responses.calls) == 1
        assert_custom_task_version_dependency_build(
            dependency_build, custom_task_version_dependency_build_server_data
        )

    @responses.activate
    def test_start_dependency_build__with_waiting(
        self,
        dependency_img_build_route,
        custom_task_version,
        custom_task_version_dependency_build_server_data,
    ):
        version = CustomTaskVersion.from_server_data(custom_task_version)
        # simulate the response of starting a dependency image build
        response_data = copy.copy(custom_task_version_dependency_build_server_data)
        responses.add(
            responses.POST,
            dependency_img_build_route,
            status=202,
            content_type="application/json",
            json=response_data,
        )
        # simulate the 1st pull of status
        response_data = copy.copy(custom_task_version_dependency_build_server_data)
        response_data["buildStatus"] = "processing"
        responses.add(
            responses.GET,
            dependency_img_build_route,
            status=200,
            content_type="application/json",
            json=response_data,
        )
        # simulate the 2nd pull of status
        response_data = copy.copy(custom_task_version_dependency_build_server_data)
        response_data["buildStatus"] = "success"
        responses.add(
            responses.GET,
            dependency_img_build_route,
            status=200,
            content_type="application/json",
            json=response_data,
        )

        with mock.patch("time.sleep"):
            dependency_build = version.start_dependency_build_and_wait(max_wait=DEFAULT_MAX_WAIT)
        assert len(responses.calls) == 3
        # check the final response
        assert_custom_task_version_dependency_build(dependency_build, response_data)

    @responses.activate
    def test_start_dependency_build__raise_timeout_error(
        self,
        dependency_img_build_route,
        custom_task_version,
        custom_task_version_dependency_build_server_data,
    ):
        version = CustomTaskVersion.from_server_data(custom_task_version)
        # simulate the response of starting a dependency image build
        response_data = custom_task_version_dependency_build_server_data
        responses.add(
            responses.POST,
            dependency_img_build_route,
            status=202,
            content_type="application/json",
            json=response_data,
        )
        # simulate the pull of status
        response_data["buildStatus"] = "processing"
        responses.add(
            responses.GET,
            dependency_img_build_route,
            status=200,
            content_type="application/json",
            json=response_data,
        )

        with mock.patch("time.sleep"):
            with pytest.raises(errors.AsyncTimeoutError):
                version.start_dependency_build_and_wait(max_wait=0)

    @responses.activate
    def test_get_dependency_build(
        self,
        dependency_img_build_route,
        custom_task_version,
        custom_task_version_dependency_build_server_data,
    ):
        version = CustomTaskVersion.from_server_data(custom_task_version)
        response_data = custom_task_version_dependency_build_server_data
        responses.add(
            responses.GET,
            dependency_img_build_route,
            status=200,
            content_type="application/json",
            json=response_data,
        )
        build_info = version.get_dependency_build()
        assert len(responses.calls) == 1
        assert_custom_task_version_dependency_build(build_info, response_data)

    @responses.activate
    @pytest.mark.parametrize("file_directory", [False, "file_directory"])
    def test_download_dependency_build_log(
        self, file_directory, dependency_img_build_log_route, custom_task_version, tmpdir,
    ):
        version = CustomTaskVersion.from_server_data(custom_task_version)
        responses.add(
            responses.GET,
            dependency_img_build_log_route,
            status=200,
            body="body",
            headers={"Content-Disposition": "attachment; filename=default_filename"},
        )

        downloaded_file = tmpdir.mkdir(file_directory) if file_directory else tmpdir
        version.download_dependency_build_log(file_directory=str(downloaded_file))
        assert len(responses.calls) == 1
        downloaded_file = (
            tmpdir.join(file_directory, "default_filename")
            if file_directory
            else tmpdir.join("default_filename")
        )
        assert downloaded_file.read() == "body"

    @responses.activate
    def test_cancel_dependency_build(self, dependency_img_build_route, custom_task_version):
        version = CustomTaskVersion.from_server_data(custom_task_version)
        responses.add(responses.DELETE, dependency_img_build_route, status=204)

        version.cancel_dependency_build()

        assert len(responses.calls) == 1



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

import pytest
import responses

from datarobot import DataStore, SharingAccess
from datarobot.utils import parse_time
from tests.utils import request_body_to_json


@responses.activate
def test_list(data_stores_endpoint, data_stores_list_server_resp):
    responses.add(responses.GET, data_stores_endpoint, json=data_stores_list_server_resp)

    data_stores = DataStore.list()

    for data_store, server_payload in zip(data_stores, data_stores_list_server_resp["data"]):
        assert data_store.id == server_payload["id"]
        assert data_store.type == server_payload["type"]
        assert data_store.canonical_name == server_payload["canonicalName"]
        assert data_store.creator == server_payload["creator"]
        assert data_store.updated == parse_time(server_payload["updated"])
        assert data_store.params.driver_id == server_payload["params"]["driverId"]
        assert data_store.params.jdbc_url == server_payload["params"]["jdbcUrl"]


@responses.activate
def test_list_no_jdbc_url(data_stores_endpoint, data_stores_list_server_resp_no_jdbc_url):
    responses.add(
        responses.GET, data_stores_endpoint, json=data_stores_list_server_resp_no_jdbc_url
    )

    data_stores = DataStore.list()

    for data_store, server_payload in zip(
        data_stores, data_stores_list_server_resp_no_jdbc_url["data"]
    ):
        assert data_store.id == server_payload["id"]
        assert data_store.type == server_payload["type"]
        assert data_store.canonical_name == server_payload["canonicalName"]
        assert data_store.creator == server_payload["creator"]
        assert data_store.updated == parse_time(server_payload["updated"])
        assert data_store.params.driver_id == server_payload["params"]["driverId"]
        assert data_store.params.jdbc_url is None
        assert "jdbcUrl" not in server_payload["params"]


@responses.activate
def test_get(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )

    data_store = DataStore.get(data_store_id)

    assert data_store.id == postgresql_data_store_server_resp["id"]
    assert data_store.type == postgresql_data_store_server_resp["type"]
    assert data_store.canonical_name == postgresql_data_store_server_resp["canonicalName"]
    assert data_store.creator == postgresql_data_store_server_resp["creator"]
    assert data_store.updated == parse_time(postgresql_data_store_server_resp["updated"])
    assert data_store.params.driver_id == postgresql_data_store_server_resp["params"]["driverId"]
    assert data_store.params.jdbc_url == postgresql_data_store_server_resp["params"]["jdbcUrl"]


@responses.activate
def test_get_no_jdbc_url(data_stores_endpoint, postgresql_data_store_server_resp_no_jdbc_url):
    data_store_id = postgresql_data_store_server_resp_no_jdbc_url["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp_no_jdbc_url,
    )

    data_store = DataStore.get(data_store_id)

    server_resp = postgresql_data_store_server_resp_no_jdbc_url

    assert data_store.id == server_resp["id"]
    assert data_store.type == server_resp["type"]
    assert data_store.canonical_name == server_resp["canonicalName"]
    assert data_store.creator == server_resp["creator"]
    assert data_store.updated == parse_time(server_resp["updated"])
    assert data_store.params.driver_id == server_resp["params"]["driverId"]
    assert data_store.params.jdbc_url is None
    assert "jdbcUrl" not in server_resp["params"]


@responses.activate
def test_future_proof(data_stores_endpoint, postgresql_data_store_server_resp):
    aug_res = dict(postgresql_data_store_server_resp, newest="NEW_KEY")
    aug_res["params"]["newKey"] = "hihihi"
    data_store_id = aug_res["id"]
    responses.add(responses.GET, "{}{}/".format(data_stores_endpoint, data_store_id), json=aug_res)

    data_store = DataStore.get(data_store_id)

    assert data_store.id == data_store_id


@responses.activate
def test_create(data_stores_endpoint, postgresql_data_store_server_resp):
    responses.add(responses.POST, data_stores_endpoint, json=postgresql_data_store_server_resp)

    data_store = DataStore.create(
        data_store_type=postgresql_data_store_server_resp["type"],
        canonical_name=postgresql_data_store_server_resp["canonicalName"],
        driver_id=postgresql_data_store_server_resp["params"]["driverId"],
        jdbc_url=postgresql_data_store_server_resp["params"]["jdbcUrl"],
    )

    assert data_store.id == postgresql_data_store_server_resp["id"]
    assert data_store.type == postgresql_data_store_server_resp["type"]
    assert data_store.canonical_name == postgresql_data_store_server_resp["canonicalName"]
    assert data_store.creator == postgresql_data_store_server_resp["creator"]
    assert data_store.updated == parse_time(postgresql_data_store_server_resp["updated"])
    assert data_store.params.driver_id == postgresql_data_store_server_resp["params"]["driverId"]
    assert data_store.params.jdbc_url == postgresql_data_store_server_resp["params"]["jdbcUrl"]


@responses.activate
def test_update(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )
    updated_data_store_server_resp = dict(
        postgresql_data_store_server_resp,
        canonicalName="updated_canonical_name",
        params={"driverId": "updated_driver_id", "jdbcUrl": "updated_jdbc_url"},
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=updated_data_store_server_resp,
    )

    data_store = DataStore.get(data_store_id)

    data_store.update(
        canonical_name=updated_data_store_server_resp["canonicalName"],
        driver_id=updated_data_store_server_resp["params"]["driverId"],
        jdbc_url=updated_data_store_server_resp["params"]["jdbcUrl"],
    )

    assert data_store.id == updated_data_store_server_resp["id"]
    assert data_store.type == updated_data_store_server_resp["type"]
    assert data_store.canonical_name == updated_data_store_server_resp["canonicalName"]
    assert data_store.creator == updated_data_store_server_resp["creator"]
    assert data_store.updated == parse_time(postgresql_data_store_server_resp["updated"])
    assert data_store.params.driver_id == updated_data_store_server_resp["params"]["driverId"]
    assert data_store.params.jdbc_url == updated_data_store_server_resp["params"]["jdbcUrl"]


@responses.activate
def test_unable_to_change_type(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )

    data_store = DataStore.get(data_store_id)
    data_store._type = "new_type"
    data_store.update()
    new_data_store = DataStore.get(data_store_id)

    assert new_data_store.type == postgresql_data_store_server_resp["type"]


@responses.activate
def test_unable_to_change_creator(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )

    data_store = DataStore.get(data_store_id)
    data_store._creator = "new_creator"
    data_store.update()
    new_data_store = DataStore.get(data_store_id)

    assert new_data_store.creator == postgresql_data_store_server_resp["creator"]


@responses.activate
def test_unable_to_change_updated_time(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )
    responses.add(
        responses.PATCH,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )

    data_store = DataStore.get(data_store_id)
    data_store._updated = datetime.now()
    data_store.update()
    new_data_store = DataStore.get(data_store_id)

    assert new_data_store.updated == parse_time(postgresql_data_store_server_resp["updated"])


@responses.activate
def test_delete(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )
    delete_url = "{}{}/".format(data_stores_endpoint, data_store_id)
    responses.add(responses.DELETE, delete_url)

    data_store = DataStore.get(data_store_id)
    data_store.delete()

    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url


@responses.activate
def test_connection_test(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )
    responses.add(
        responses.POST,
        "{}{}/test/".format(data_stores_endpoint, data_store_id),
        json={"message": "Connection successful"},
    )

    data_store = DataStore.get(data_store_id)
    resp = data_store.test(username="my_username", password="my_password")

    assert resp["message"] == "Connection successful"


@responses.activate
def test_schemas_list(
    data_stores_endpoint, postgresql_data_store_server_resp, postgresql_schemas_list_server_resp
):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )
    responses.add(
        responses.POST,
        "{}{}/schemas/".format(data_stores_endpoint, data_store_id),
        json=postgresql_schemas_list_server_resp,
    )

    data_store = DataStore.get(data_store_id)
    resp = data_store.schemas(username="my_username", password="my_password")

    assert resp == postgresql_schemas_list_server_resp


@responses.activate
def test_tables_list(
    data_stores_endpoint, postgresql_data_store_server_resp, postgresql_tables_list_server_resp
):
    data_store_id = postgresql_data_store_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp,
    )
    responses.add(
        responses.POST,
        "{}{}/tables/".format(data_stores_endpoint, data_store_id),
        json=postgresql_tables_list_server_resp,
    )

    data_store = DataStore.get(data_store_id)
    resp = data_store.tables(username="my_username", password="my_password", schema="demo")

    assert resp == postgresql_tables_list_server_resp


@responses.activate
@pytest.mark.parametrize("extra_keys", [{}, {"newKey": "blah"}])
def test_get_access_list(data_stores_endpoint, postgresql_data_store_server_resp, extra_keys):
    data_store = DataStore.from_server_data(postgresql_data_store_server_resp)
    access_list = {
        "data": [
            {
                "username": "me@datarobot.com",
                "userId": "1234deadbeeffeeddead4321",
                "role": "OWNER",
                "canShare": True,
            }
        ],
        "count": 1,
        "previous": None,
        "next": None,
    }
    access_record = access_list["data"][0]
    access_record.update(extra_keys)
    responses.add(
        responses.GET,
        "{}{}/accessControl/".format(data_stores_endpoint, data_store.id),
        json=access_list,
    )

    response = data_store.get_access_list()
    assert len(response) == 1

    share_info = response[0]
    assert share_info.username == access_record["username"]
    assert share_info.user_id == access_record["userId"]
    assert share_info.role == access_record["role"]
    assert share_info.can_share == access_record["canShare"]


@responses.activate
def test_share(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store = DataStore.from_server_data(postgresql_data_store_server_resp)
    request = SharingAccess("me@datarobot.com", "EDITOR")
    responses.add(
        responses.PATCH,
        "{}{}/accessControl/".format(data_stores_endpoint, data_store.id),
        status=204,
    )

    data_store.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"data": [{"username": "me@datarobot.com", "role": "EDITOR"}]}


@responses.activate
def test_remove(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store = DataStore.from_server_data(postgresql_data_store_server_resp)
    request = SharingAccess("me@datarobot.com", None)
    responses.add(
        responses.PATCH,
        "{}{}/accessControl/".format(data_stores_endpoint, data_store.id),
        status=204,
    )

    data_store.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"data": [{"username": "me@datarobot.com", "role": None}]}


@responses.activate
def test_share_with_grant(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store = DataStore.from_server_data(postgresql_data_store_server_resp)
    responses.add(
        responses.PATCH,
        "{}{}/accessControl/".format(data_stores_endpoint, data_store.id),
        status=204,
    )

    data_store.share(
        [
            SharingAccess("me@datarobot.com", "EDITOR", can_share=True),
            SharingAccess("other@datarobot.com", "CONSUMER", can_share=False),
        ]
    )

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {
        "data": [
            {"username": "me@datarobot.com", "role": "EDITOR", "canShare": True},
            {"username": "other@datarobot.com", "role": "CONSUMER", "canShare": False},
        ]
    }



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

from numpy import isclose
import pytest
import responses

from datarobot import Model
from datarobot.models import ModelParameters


@pytest.fixture
def model_parameters_data():
    return {
        "parameters": [
            {"name": u"Intercept", "value": 1},
            {"name": u"Link function", "value": u"logit"},
        ],
        "derivedFeatures": [
            {
                "coefficient": -0.015,
                "originalFeature": u"A1Cresult",
                "derivedFeature": u"A1Cresult->7",
                "type": u"CAT",
                "transformations": [{"name": u"One-hot", "value": u"'>7'"}],
            },
            {
                "coefficient": 0.014,
                "originalFeature": u"A1Cresult",
                "derivedFeature": u"A1Cresult->8",
                "type": u"CAT",
                "transformations": [{"name": u"One-hot", "value": u"'>8'"}],
            },
            {
                "coefficient": -0.133,
                "originalFeature": u"A1Cresult",
                "derivedFeature": u"A1Cresult-Norm",
                "type": u"CAT",
                "transformations": [{"name": u"One-hot", "value": u"'Norm'"}],
            },
            {
                "coefficient": 0.40552547397855254,
                "type": u"NUM",
                "derivedFeature": u"STANDARDIZED_revol_util",
                "originalFeature": u"revol_util",
                "transformations": [
                    {"name": u"Missing imputation", "value": 47.6},
                    {"name": u"Standardize", "value": u"(47.781034375,27.955074618)"},
                ],
            },
        ],
    }


@pytest.fixture
def freq_sev_model_parameters_data(model_parameters_data):
    freq_sev_data = model_parameters_data.copy()
    for derived_feature in freq_sev_data["derivedFeatures"]:
        derived_feature["stageCoefficients"] = [
            {"stage": "frequency", "coefficient": 1.0},
            {"stage": "severity", "coefficient": 2.0},
        ]
    return freq_sev_data


@pytest.fixture
def broken_model_parameters_data(model_parameters_data):
    broken_data = model_parameters_data.copy()
    bad_data = {"some_broken_field": "some_broken_value"}
    broken_data["parameters"][0].update(bad_data)
    broken_data["derivedFeatures"][0].update(bad_data)
    broken_data.update(bad_data)
    return broken_data


@pytest.fixture
def model_parameters_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/parameters/".format(project_id, model_id)


@responses.activate
def test_instantiation(model_parameters_url, model_parameters_data, project_id, model_id):
    responses.add(
        responses.GET,
        model_parameters_url,
        status=200,
        content_type="application/json",
        body=json.dumps(model_parameters_data),
    )

    mp = ModelParameters.get(project_id, model_id)

    params = mp.parameters
    expected_params = model_parameters_data["parameters"]
    for i in range(len(params)):
        assert params[i]["name"] == expected_params[i]["name"]
        assert params[i]["value"] == expected_params[i]["value"]

    derived = mp.derived_features
    expected_derived = model_parameters_data["derivedFeatures"]
    for i in range(len(derived)):
        assert isclose(derived[i]["coefficient"], expected_derived[i]["coefficient"])
        assert derived[i]["original_feature"] == expected_derived[i]["originalFeature"]
        assert derived[i]["derived_feature"] == expected_derived[i]["derivedFeature"]
        assert derived[i]["type"] == expected_derived[i]["type"]
        assert derived[i]["transformations"] == expected_derived[i]["transformations"]


@responses.activate
def test_freq_sev_parameters(
    model_parameters_url, freq_sev_model_parameters_data, project_id, model_id
):
    responses.add(
        responses.GET,
        model_parameters_url,
        status=200,
        content_type="application/json",
        body=json.dumps(freq_sev_model_parameters_data),
    )

    mp = ModelParameters.get(project_id, model_id)

    derived_features = mp.derived_features
    expected_derived_features = freq_sev_model_parameters_data["derivedFeatures"]
    for i in range(len(derived_features)):
        assert len(derived_features[i]["stage_coefficients"]) == 2
        for j in range(len(derived_features[i]["stage_coefficients"])):
            assert isclose(
                derived_features[i]["stage_coefficients"][j]["coefficient"],
                expected_derived_features[i]["stageCoefficients"][j]["coefficient"],
            )


@responses.activate
def test_future_proof(model_parameters_url, broken_model_parameters_data, project_id, model_id):
    responses.add(
        responses.GET,
        model_parameters_url,
        status=200,
        content_type="application/json",
        body=json.dumps(broken_model_parameters_data),
    )

    mp = ModelParameters.get(project_id, model_id)

    assert len(mp.parameters) == 2
    for i in range(len(mp.parameters)):
        assert "some_broken_field" not in mp.parameters[i]

    assert len(mp.derived_features) == 4
    for i in range(len(mp.derived_features)):
        assert "some_broken_field" not in mp.derived_features[i]


@responses.activate
def test_get_parameters(model_parameters_url, model_parameters_data, project_id, model_id):
    responses.add(
        responses.GET,
        model_parameters_url,
        status=200,
        content_type="application/json",
        body=json.dumps(model_parameters_data),
    )

    mp = ModelParameters.get(project_id, model_id)
    model = Model(id=model_id, project_id=project_id)
    assert mp.parameters == model.get_parameters().parameters
